# -*- coding: utf-8 -*-
"""
This module contains functionality useful for displaying an eye tracker
independent validation process that tests the accuracy of the eye tracker
being used. The functionality could also be used to drive calibration graphics
for eye tracker models that require psychopy to perform the stimulus
presentation during calibration.

To use this module, the following high level steps are generally preformed:
* Create an instance of TargetStim, specifying the fixation target appearance.
* Create an instance of PositionGrid, which defines target position information.
* Create a TargetPosSequenceStim instance, providing the TargetStim and
  PositionGrid objects created, as well as the Trigger's which should be used
  to transition from one target position to another during the sequence of
  target graphics presentation and the defined positions.
* Use TargetPosSequenceStim.display() to run the full presentation procedure.
* Use TargetPosSequenceStim.targetdata to access information about each target
  position displayed and the events collected during the display duration for
  each position.
"""
from __future__ import division, print_function, absolute_import

from weakref import proxy
import numpy as np
from time import sleep
import os
import sys
from PIL import Image
from collections import OrderedDict

from psychopy import visual, core
from psychopy.iohub.util import win32MessagePump, normjoin
from psychopy.iohub.constants import EventConstants
from psychopy.iohub.client import ioHubConnection

from .visualUtil import Trigger, TimeTrigger, DeviceEventTrigger

getTime = core.getTime


class TargetStim(object):
    """TargetStim creates a target graphic that can be used during a sequential
    fixation task.

    The stim consists of two psychopy circle stim, one is used to draw
    the outer edge of the target and the target body fill. The second is
    used to draw a dot in the center of the larger circle.

    """

    def __init__(self,
                 win,
                 radius=None,       # The outer radius of the target.
                 fillcolor=None,    # The color used to fill the target body.
                 edgecolor=None,    # The color for the edge around the target.
                 edgewidth=None,    # The thickness of the target outer edge.
                 dotcolor=None,     # The color of the central target dot.
                 dotradius=None,    # The radius to use for the target dot.
                 units=None,
                 # The psychopy unit type of any size values.
                 colorspace=None,   # The psychopy color space of any colors.
                 opacity=1.0,
                 # The transparency of the target (0.0 - 1.0)
                 contrast=1.0       # The contrast of the target stim.
                 ):
        self.win = proxy(win)
        self.stim = []
        self.radius = radius
        outer = visual.Circle(self.win,
                              radius=radius,
                              fillColor=fillcolor,
                              lineColor=edgecolor,
                              lineWidth=edgewidth,
                              edges=int(np.pi * radius),
                              units=units,
                              lineColorSpace=colorspace,
                              fillColorSpace=colorspace,
                              opacity=opacity,
                              contrast=contrast,
                              interpolate=True,
                              autoLog=False)
        self.stim.append(outer)

        if dotcolor and dotcolor != fillcolor:
            centerdot = visual.Circle(self.win,
                                      radius=dotradius,
                                      fillColor=dotcolor,
                                      lineColor=dotcolor,
                                      lineWidth=0.0,
                                      edges=int(np.pi * dotradius),
                                      interpolate=True,
                                      units=units,
                                      lineColorSpace=colorspace,
                                      fillColorSpace=colorspace,
                                      opacity=opacity,
                                      contrast=contrast,
                                      autoLog=False)
            self.stim.append(centerdot)

    def setRadius(self, r):
        """Update the radius of the target stim."""
        self.stim[0].radius = r

    def setPos(self, pos):
        """Set the position of the target stim.

        The target center will be drawn at the position given.

        """
        for s in self.stim:
            s.setPos(pos)

    def draw(self):
        """Draw the Target stim.

        (this simply calls the draw method of the psychopy stim used, in
        the correct order).

        """
        for s in self.stim:
            s.draw()

    def contains(self, p):
        return self.stim[0].contains(p)


class PositionGrid(object):
    """PositionGrid provides a flexible way to generate a set of x,y position
    values within the boundaries of the psychopy window object provided.

    The class provides a set of arguments that represent commonly needed
    constrains when creating a position list for a sequential fixation type
    task. This allows for a wide variety of position lists to be generated.

    PositionGrid supports the len() function, and returns the number of
    positions generated based on the supplied parameters. If repeatfirstpos
    is true, len(posgrid) == number of unique positions + 1 (a repeat of the
    first position value).

    PositionGrid is a generator, so the normal way to access the positions from
    the class is to use a for loop or with statement:

    posgrid = PositionGrid(....)
    for pos in posgrid:
        # do something cool with the pos

    """

    def __init__(self,
                 winSize=None,       # Window width,height bounds.
                 shape=None,         # Defines the number of columns and rows of
                 # positions needed. If shape is an array of
                 # two elements, it defines the col,row shape
                 # for position layout. Position count will
                 # equal rows*cols. If shape is a single
                 # int, the position grid col,row shape will
                 # be shape x shape.
                 posCount=None,      # Defines the number of positions to create
                 # without any col,row position constraint.
                 leftMargin=None,    # Specify the minimum valid horz position.
                 rightMargin=None,   # Limit horz positions to be < max horz
                 # position minus rightMargin.
                 topMargin=None,     # Limit vert positions to be < max vert
                 # position minus topMargin.
                 bottomMargin=None,  # Specify the minimum valid vert position.
                 scale=1.0,          # Scale can be one or two numbers, each
                 # between 0.0 and 1.0. If a tuple is
                 # provided, it represents the horz, vert
                 # scale to be applied to window width,
                 # height. If a single number is
                 # given, the same scale will be applied to
                 # both window width and height. The scaled
                 # window size is centered on the original
                 # window size to define valid position area.
                 posList=None,       # Provide an existing list of (x,y)
                 # positions. If posList is provided, the
                 # shape, posCount, margin and scale arg's
                 # are ignored.
                 noiseStd=None,      # Add a random shift to each position based
                 # on a normal distribution with mean = 0.0
                 # and sigma equal to noiseStd. Specify
                 # value based on units being used.
                 firstposindex=0,    # Specify which position in the position
                 # list should be displayed first. This
                 # position is not effected by randomization.
                 repeatfirstpos=True  # If the first position in the list should
                 # be provided as the last position as well,
                 # set to True. In this case, the number of
                 # positions returned will be position
                 # count + 1. False indicated the first
                 # position should not be repeated.
                 ):

        self.posIndex = 0
        self.positions = None
        self.posOffsets = None

        self.winSize = winSize
        self.firstposindex = firstposindex

        self.repeatfirstpos = repeatfirstpos

        self.horzStd, self.vertStd = None, None
        if noiseStd:
            if hasattr(noiseStd, '__len__'):
                self.horzStd, self.vertStd = noiseStd
            else:
                self.horzStd, self.vertStd = noiseStd, noiseStd

        horzScale, vertScale = None, None
        if scale:
            if hasattr(scale, '__len__'):
                horzScale, vertScale = scale
            else:
                horzScale, vertScale = scale, scale

        rowCount, colCount = None, None
        if shape:
            if hasattr(shape, '__len__'):
                colCount, rowCount = shape
            else:
                rowCount, colCount = shape, shape

        if posList:
            if (len(posList) == 2 and
                    len(posList[0]) != 2 and
                    len(posList[0]) == len(posList[1])):
                # positions were provided in ((x1,x2,..,xn),(y1,y2,..,yn))
                # format
                self.positions = np.column_stack((posList[0],
                                                  posList[1]))
            elif len(posList[0]) == 2:
                self.positions = np.asarray(posList)
            else:
                raise ValueError('PositionGrid posList kwarg must be in'
                                 ' ((x1,y1),(x2,y2),..,(xn,yn))'
                                 ' or ((x1,x2,..,xn),(y1,y2,..,yn)) format')

        if self.positions is None and (posCount or (rowCount and colCount)):
            # create posCount random grid positions within winSize
            if winSize is not None:
                pixw, pixh = winSize
                xmin = 0.0
                xmax = 1.0
                ymin = 0.0
                ymax = 1.0

                if leftMargin:
                    if leftMargin < pixw:
                        xmin = leftMargin / pixw
                    else:
                        raise ValueError(
                            'PositionGrid leftMargin kwarg must be'
                            ' < winSize[0]')
                if rightMargin:
                    if rightMargin < pixw:
                        xmax = 1.0 - rightMargin / pixw
                    else:
                        raise ValueError(
                            'PositionGrid rightMargin kwarg must be'
                            ' < winSize[0]')
                if topMargin:
                    if topMargin < pixh:
                        ymax = 1.0 - topMargin / pixh
                    else:
                        raise ValueError('PositionGrid topMargin kwarg must be'
                                         ' < winSize[1]')
                if bottomMargin:
                    if bottomMargin < pixh:
                        ymin = bottomMargin / pixh
                    else:
                        raise ValueError(
                            'PositionGrid bottomMargin kwarg must be'
                            ' < winSize[1]')
                        ymin = bottomMargin / pixh

                if horzScale:
                    if 0.0 < horzScale <= 1.0:
                        xmin += (1.0 - horzScale) / 2.0
                        xmax -= (1.0 - horzScale) / 2.0
                else:
                    raise ValueError('PositionGrid horzScale kwarg must be'
                                     ' 0.0 > horzScale <= 1.0')

                if vertScale:
                    if 0.0 < vertScale <= 1.0:
                        ymin += (1.0 - vertScale) / 2.0
                        ymax -= (1.0 - vertScale) / 2.0
                else:
                    raise ValueError('PositionGrid vertScale kwarg must be'
                                     ' 0.0 > vertScale <= 1.0')
                if posCount:
                    colCount = int(np.sqrt(posCount))
                    rowCount = colCount
                    xps = np.random.uniform(
                        xmin, xmax, colCount) * pixw - pixw / 2.0
                    yps = np.random.uniform(
                        ymin, ymax, rowCount) * pixh - pixh / 2.0
                else:
                    xps = np.linspace(xmin, xmax, colCount) * pixw - pixw / 2.0
                    yps = np.linspace(ymin, ymax, rowCount) * pixh - pixh / 2.0

                xps, yps = np.meshgrid(xps, yps)
                self.positions = np.column_stack(
                    (xps.flatten(), yps.flatten()))

            else:
                raise ValueError('PositionGrid posCount kwarg also requires'
                                 'winSize to be provided.')

        if self.positions is None:
            raise AttributeError('PositionGrid is unable to generate positions'
                                 'based on the provided kwargs.')

        if self.firstposindex:
            fpos = self.positions[self.firstposindex]
            self.positions = np.delete(self.positions, self.firstposindex, 0)
            self.positions = np.insert(self.positions, 0, fpos, 0)

        self._generatePosOffsets()

    def __len__(self):
        if self.repeatfirstpos:
            return len(self.positions) + 1
        else:
            return len(self.positions)

    def plot(self, **kwargs):
        """Uses Matplotlib to create a figure illustrating the screen
        positions, and presentation order, for the PositionGrid's current
        state.

        **kwargs will be directly passed to pyplot.scatter() as kwargs.

        """
        from matplotlib import pyplot as pl
        x = [p[0] for p in self]
        y = [p[1] for p in self]
        pixw, pixh = self.winSize
        pl.clf()
        pl.scatter(x, y, **kwargs)
        pl.xlim(-pixw / 2, pixw / 2)
        pl.ylim(-pixh / 2, pixh / 2)
        for i in range(len(x)):
            pl.text(x[i], y[i], str(i), size=11, horizontalalignment='center',
                    verticalalignment='center')
        pl.xlabel('Horizontal Target Position')
        pl.ylabel('Vertical Target Position')
        pl.title(
            'PositionGrid Generated Screen Locations\n(Point Number is List Index)')
        pl.show()

    def randomize(self):
        """
        Randomize the positions within the position list. If a first position
        index was been provided, randomization only occurs for positions[1:].

        This can be called multiple times if the same position list is being used
        repeatedly and a random presentation order is needed.

        Each time randomize() is called, if noiseStd is != 0, a new set of
        normally distributed offsets are created for the target positions.
        """
        if not self.firstposindex:
            np.random.shuffle(self.positions)
        else:
            firstpos = self.positions[0]
            self.positions = np.delete(self.positions, 0, 0)
            np.random.shuffle(self.positions)
            self.positions = np.insert(self.positions, 0, firstpos, 0)
        self._generatePosOffsets()

    def _generatePosOffsets(self):
        """Create a new set of position displayment 'noise' based on the
        noiseStd value given when the object was initialized."""
        horzPosOffsetList = np.zeros((len(self), 1))
        if self.horzStd:
            horzPosOffsetList = np.random.normal(0.0, self.horzStd,
                                                 len(self))
        vertPosOffsetList = np.zeros((len(self), 1))
        if self.vertStd:
            vertPosOffsetList = np.random.normal(0.0, self.vertStd,
                                                 len(self))
        self.posOffsets = np.column_stack(
            (vertPosOffsetList, horzPosOffsetList))

    def __iter__(self):
        return self

    # Python 3 compatibility
    def __next__(self):
        return self.next()

    def next(self):
        """Returns the next position in the list. Usually this method is not
        called directly. Instead, positions are accessed by iterating over the
        PositionGrid object.

        pos = PositionGrid(....)

        for p in pos:
            # do something cool with it
            pass

        """
        if self.posIndex < len(self.positions):
            pos = self.positions[self.posIndex] + \
                self.posOffsets[self.posIndex]
            self.posIndex = self.posIndex + 1
            return pos
        elif self.repeatfirstpos and self.posIndex == len(self.positions):
            pos = self.positions[0] + self.posOffsets[0]
            self.posIndex = self.posIndex + 1
            return pos
        else:
            self.posIndex = 0
            raise StopIteration()


class TargetPosSequenceStim(object):
    """TargetPosSequenceStim combines an instance of a Target stim and an
    instance of a PositionGrid to create everything needed to present the
    target at each position returned by the PositionGrid instance within the
    psychopy window used to create the Target stim. The target is presented at
    each position sequentially.

    By providing keyword arguments to the TargetPosSequenceStim.display(...)
    method, position animation between target positions, and target stim
    expansion and / or contraction transitions are possible.

    psychopy.iohub.Trigger based classes are used to define the criteria used to
    start displaying the next target position graphics. By providing a list
    of a TimerTrigger and a set of DeviceEventTriggers, complex criteria
    for target position pacing can be easily defined for use during the display
    period.

    iohub devices can be provided in the storeeventsfor keyword argument.
    Events which occur during each target position presentation period are
    stored and are available at the end of the display() period, grouped by
    position index and device event types.

    """
    TARGET_STATIONARY = 1
    TARGET_MOVING = 2
    TARGET_EXPANDING = 4
    TARGET_CONTRACTING = 8

    # Experiment Message text field types and tokens
    #
    message_types = dict(
        # position_count
        BEGIN_SEQUENCE=('BEGIN_SEQUENCE', '', int),
        # position_count
        DONE_SEQUENCE=('DONE_SEQUENCE', '', int),
        # event_type_id event_time
        NEXT_POS_TRIG=('NEXT_POS_TRIG', '', int, float),
        # position_count from_x,from_y to_x,to_y
        START_DRAW=('START_DRAW', ',', int, float, float, float, float),
        # position_count from_x,from_y to_x,to_y
        SYNCTIME=('SYNCTIME', ',', int, float, float, float, float),
        # current_radius original_radius
        EXPAND_SIZE=('EXPAND_SIZE', '', float, float),
        # current_radius original_radius
        CONTRACT_SIZE=('CONTRACT_SIZE', '', float, float),
        # current_x,current_y
        POS_UPDATE=('POS_UPDATE', ',', float, float),
        # final_x,final_y
        TARGET_POS=('TARGET_POS', ',', float, float)
    )
    max_msg_type_length = max([len(s) for s in message_types])

    binocular_sample_message_element = [
        ('targ_pos_ix', np.int),
        ('last_msg_time', np.float32),
        ('last_msg_type', '|S%d'%(max_msg_type_length)),
        ('next_msg_time', np.float32),
        ('next_msg_type', '|S%d'%(max_msg_type_length)),
        ('targ_pos_x', np.float32),
        ('targ_pos_y', np.float32),
        ('targ_state', np.int),
        ('eye_time', np.float32),
        ('eye_status', np.int),
        ('left_eye_x', np.float32),
        ('left_eye_y', np.float32),
        ('left_pupil_size', np.float32),
        ('right_eye_x', np.float32),
        ('right_eye_y', np.float32),
        ('right_pupil_size', np.float32)
    ]

    monocular_sample_message_element = [
        ('targ_pos_ix', np.int),
        ('last_msg_time', np.float32),
        ('last_msg_type', '|S%d'%(max_msg_type_length)),
        ('next_msg_time', np.float32),
        ('next_msg_type', '|S%d'%(max_msg_type_length)),
        ('targ_pos_x', np.float32),
        ('targ_pos_y', np.float32),
        ('targ_state', np.int),
        ('eye_time', np.float32),
        ('eye_status', np.int),
        ('eye_x', np.float32),
        ('eye_y', np.float32),
        ('pupil_size', np.float32)
    ]

    def __init__(self,
                 win,
                 target,            # The TargetStim instance to use.
                 positions,         # The PositionGrid instance to use.
                 background=None,   # Window background color to use (if any).
                 storeeventsfor=[],  # List of iohub device objects to track
                 # events for.
                 triggers=None,     # The triggers to use for controlling target
                                    # position progression.
                 msgcategory='',     # As the display() process is preformed,
                                    # iohub experiment messages are generated
                                    # providing information on the state of the
                                    # target and position. msgcategory can be
                                    # used to define the category string to
                                    # assign to each message.
                 config=None,       # A config for all necessary kwargs. If this
                                    # is provided, all other kwargs but io
                                    # can be skipped.
                 io=None            # The ioHubConnection instance to use.
                 ):
        self.win = win
        self.target = target
        self.background = background
        self.positions = positions
        self.storeevents = storeeventsfor
        self.msgcategory = msgcategory

        if io is None:
            io = ioHubConnection.getActiveConnection()
        self.io = io

        # If storeevents is True, targetdata will be a list of dict's.
        # Each dict, among other things, contains all ioHub events that occurred
        # from when a target was first presented at a position, to when the
        # the wait period completed for that position.
        #
        self.targetdata = []
        self.triggers = None

        # Handle different valid trigger object types
        if isinstance(triggers, (list, tuple)):
            # Support is provided for a list of Trigger objects or a list of
            # strings.
            t1 = triggers[0]
            if isinstance(t1, basestring):
                # triggers is a list of strings, so try and create a list of
                # DeviceEventTrigger's using keyboard device, KEYBOARD_RELEASE
                # event type, and the triggers list elements each as the
                # event.key.
                kbtriggers = []
                kbdevice = io.getDevice('keyboard')
                KEYBOARD_RELEASE = EventConstants.KEYBOARD_RELEASE
                for c in triggers:
                    kbtriggers.append(
                        DeviceEventTrigger(
                            kbdevice,
                            event_type=KEYBOARD_RELEASE,
                            event_attribute_conditions={
                                'char': c}))
                    self.triggers = kbtriggers
            else:
                # Assume triggers is a list of Trigger objects
                self.triggers = triggers
        elif isinstance(triggers, (int, float, long)):
            # triggers is a number, so assume a TimeTrigger is wanted where
            # the delay == triggers. start time will be the fliptime of the
            # last update for drawing to the new target position.
            self.triggers = (TimeTrigger(start_time=None, delay=triggers),)
        elif isinstance(triggers, basestring):
            # triggers is a string, so try and create a
            # DeviceEventTrigger using keyboard device, KEYBOARD_RELEASE
            # event type, and triggers as the event.key.
            self.triggers = (
                DeviceEventTrigger(
                    io.getDevice('keyboard'),
                    event_type=EventConstants.KEYBOARD_RELEASE,
                    event_attribute_conditions={
                        'char': triggers}),
            )
        elif isinstance(triggers, Trigger):
            # A single Trigger object was provided
            self.triggers = (triggers,)
        else:
            raise ValueError('The triggers kwarg could not be understood as a '
                             'valid triggers input value.')

    def getIO(self):
        """Get the active ioHubConnection instance."""
        return self.io

    def _draw(self):
        """Fill the window with the specified background color and draw the
        target stim."""
        if self.background:
            self.background.draw()
        self.target.draw()

    def _animateTarget(self, topos, frompos, **kwargs):
        """Internal method.

        Any logic related to drawing the target at the new screen position,
        including any intermediate animation effects, is done here.

        Return the flip time when the target was first drawn at the newpos
        location.

        """
        io = self.getIO()
        if frompos is not None:
            velocity = kwargs.get('velocity')
            if velocity:
                starttime = getTime()
                a, b = np.abs(topos - frompos)**2
                duration = np.sqrt(a + b) / velocity
                arrivetime = duration + starttime
                fliptime = starttime
                while fliptime < arrivetime:
                    mu = (fliptime - starttime) / duration
                    tpos = frompos * (1.0 - mu) + topos * mu
                    self.target.setPos(frompos * (1.0 - mu) + topos * mu)
                    self._draw()
                    fliptime = self.win.flip()
                    io.sendMessageEvent('POS_UPDATE %.2f,%.2f' % (
                                        tpos[0], tpos[1]), self.msgcategory,
                                        sec_time=fliptime)
                    self._addDeviceEvents()

        self.target.setPos(topos)
        self._draw()
        fliptime = self.win.flip()
        io.sendMessageEvent('TARGET_POS %.2f,%.2f' % (topos[0], topos[1]),
                            self.msgcategory, sec_time=fliptime)
        self._addDeviceEvents()

        expandedscale = kwargs.get('expandedscale')
        expansionduration = kwargs.get('expansionduration')
        contractionduration = kwargs.get('contractionduration')

        initialradius = self.target.radius
        if expandedscale:
            expandedradius = self.target.radius * expandedscale

            if expansionduration:
                starttime = fliptime
                expandedtime = fliptime + expansionduration
                while fliptime < expandedtime:
                    mu = (fliptime - starttime) / expansionduration
                    cradius = initialradius * (1.0 - mu) + expandedradius * mu
                    self.target.setRadius(cradius)
                    self._draw()
                    fliptime = self.win.flip()
                    io.sendMessageEvent('EXPAND_SIZE %.2f %.2f' % (
                                        cradius, initialradius),
                                        self.msgcategory, sec_time=fliptime)
                    self._addDeviceEvents()

            if contractionduration:
                starttime = fliptime
                contractedtime = fliptime + contractionduration
                while fliptime < contractedtime:
                    mu = (fliptime - starttime) / contractionduration
                    cradius = expandedradius * (1.0 - mu) + initialradius * mu
                    self.target.setRadius(cradius)
                    self._draw()
                    fliptime = self.win.flip()
                    io.sendMessageEvent('CONTRACT_SIZE %.2f %.2f' % (
                                        cradius, initialradius),
                                        self.msgcategory, sec_time=fliptime)
                    self._addDeviceEvents()

        self.target.setRadius(initialradius)
        return fliptime

    def moveTo(self, topos, frompos, **kwargs):
        """Indicates that the target should be moved frompos to topos.

        If a PositionGrid has been provided, moveTo should not be called
        directly. Instead, use the display() method to start the full
        target position presentation sequence.

        """
        io = self.getIO()
        fpx, fpy = -1, -1
        if frompos is not None:
            fpx, fpy = frompos[0], frompos[1]
        io.sendMessageEvent('START_DRAW %d %.2f,%.2f %.2f,%.2f' % (
            self.positions.posIndex, fpx, fpy,
            topos[0], topos[1]), self.msgcategory)

        fliptime = self._animateTarget(topos, frompos, **kwargs)

        io.sendMessageEvent('SYNCTIME %d %.2f,%.2f %.2f,%.2f' % (
            self.positions.posIndex, fpx, fpy,
            topos[0], topos[1]), self.msgcategory, sec_time=fliptime)

        # wait for trigger to fire
        last_pump_time = fliptime
        while not self._hasTriggerFired(start_time=fliptime):
            if getTime() - last_pump_time >= 0.250:
                win32MessagePump()
                last_pump_time = getTime()
            sleep(0.001)

    def _hasTriggerFired(self, **kwargs):
        """Used internally to know when one of the triggers has occurred and
        the target should move to the next target position."""
        # wait for trigger to fire
        triggered = None
        for trig in self.triggers:
            if trig.triggered(**kwargs):
                triggered = trig
                break
        self._addDeviceEvents(trig.clearEventHistory(True))
        if triggered:
            # assume it was a timer trigger,so use 255 as 'event type'
            event_type_id = 255
            trig_evt = triggered.getTriggeringEvent()
            if hasattr(trig_evt, 'type'):
                # actually it was a device event trigger
                event_type_id = trig_evt.type
            # get time trigger of trigger event
            event_time = triggered.getTriggeringTime()
            self.getIO().sendMessageEvent('NEXT_POS_TRIG %d %.3f' % (
                event_type_id, event_time), self.msgcategory)
            for trig in self.triggers:
                trig.resetTrigger()
        return triggered

    def _initTargetData(self, frompos, topos):
        """Internally used to create the data structure used to store position
        information and events which occurred during each target position
        period."""
        if self.storeevents:
            deviceevents = {}
            for device in self.storeevents:
                deviceevents[device] = []
        self.targetdata.append(dict(frompos=frompos,
                                    topos=topos,
                                    events=deviceevents
                                    )
                               )

    def _addDeviceEvents(self, device_event_dict={}):
        dev_event_buffer = self.targetdata[-1]['events']
        for dev, dev_events in dev_event_buffer.items():
            if dev in device_event_dict:
                dev_events.extend(device_event_dict[dev])
            else:
                dev_events.extend(dev.getEvents())

    def display(self, **kwargs):
        """Display the target at each point in the position grid, performing
        target animation if requested. The target then holds position until one
        of the specified triggers occurs, resulting in the target moving to the
        next position in the positiongrid.

        To setup target animation between grid positions, the following keyword
        arguments are supported. If an option is not specified, the animation
        related to it is not preformed.

        velocity: The rate (units / second) at which the target should move
                  from a current target position to the next target position.
                  The value should be in the unit type the target stimulus
                  is using.

        expandedscale: When a target stimulus is at the current grid position,
                       the target graphic can expand to a size equal to the
                       original target radius * expandedscale.

        expansionduration: If expandedscale has been specified, this option is
                           used to set how long it should take for the target to
                           reach the full expanded target size. Time is in sec.

        contractionduration: If a target has been expanded, this option is used
                             to specify how many seconds it should take for the
                             target to contract back to the original target
                             radius.

        Note that target expansion and contraction change the target stimulus
        outer diameter only. The edge thickness and central dot radius do not
        change.

        All movement and size changes are linear in fashion.

        For example, to display a static target at each grid position::

        targetsequence.display()

        To have the target stim move between each grid position
        at 400 pixels / sec and not expand or contract::

        targetsequence.display(velocity=400.0)

        If the target should jump from one grid position to the next, and then
        expand to twice the radius over a 0.5 second period::

        targetsequence.display(
                                expandedscale=2.0,
                                expansionduration=0.50
                              )

        To do a similar animation as the previous example, but also have the
        target contract back to it's original size over 0.75 seconds::

        targetsequence.display(
                                expandedscale=2.0,
                                expansionduration=0.50,
                                contractionduration=0.75
                              )

        When this method returns, the target has been displayed at all
        positions. Data collected for each position period can be accessed via
        the targetdata attribute.

        """
        del self.targetdata[:]
        prevpos = None

        io = self.getIO()
        io.clearEvents('all')
        io.sendMessageEvent('BEGIN_SEQUENCE {0}'.format(
            len(self.positions.positions)), self.msgcategory)
        turn_rec_off = []
        for d in self.storeevents:
            if not d.isReportingEvents():
                d.enableEventReporting(True)
                turn_rec_off.append(d)

        sleep(0.025)
        for pos in self.positions:
            self._initTargetData(prevpos, pos)
            self._addDeviceEvents()
            self.moveTo(pos, prevpos, **kwargs)
            prevpos = pos
            self._addDeviceEvents()

        for d in turn_rec_off:
            d.enableEventReporting(False)

        io.sendMessageEvent('DONE_SEQUENCE {0}'.format(
            len(self.positions.positions)), self.msgcategory)
        sleep(0.025)
        self._addDeviceEvents()
        io.clearEvents('all')

    def _processMessageEvents(self):
        self.target_pos_msgs = []
        self.saved_pos_samples = []
        for pd in self.targetdata:
            frompos = pd.get('frompos')
            topos = pd.get('topos')
            events = pd.get('events')

            # create a dict of device labels as keys, device events as value
            devlabel_events = {}
            for k, v in events.items():
                devlabel_events[k.getName()] = v

            samples = devlabel_events.get('tracker', [])
            # remove any eyetracker events that are not samples
            samples = [
                s for s in samples if s.type in (
                    EventConstants.BINOCULAR_EYE_SAMPLE,
                    EventConstants.MONOCULAR_EYE_SAMPLE)]
            self.saved_pos_samples.append(samples)

            self.sample_type = self.saved_pos_samples[0][0].type
            if self.sample_type == EventConstants.BINOCULAR_EYE_SAMPLE:
                self.sample_msg_dtype = self.binocular_sample_message_element
            else:
                self.sample_msg_dtype = self.monocular_sample_message_element

            messages = devlabel_events.get('experiment', [])
            msg_lists = []
            for m in messages:
                temp = m.text.strip().split()
                msg_type = self.message_types.get(temp[0])
                if msg_type:
                    current_msg = [m.time, m.category]
                    if msg_type[1] == ',':
                        for t in temp:
                            current_msg.extend(t.split(','))
                    else:
                        current_msg.extend(temp)

                    for mi, dtype in enumerate(msg_type[2:]):
                        current_msg[mi + 3] = dtype(current_msg[mi + 3])

                    msg_lists.append(current_msg)

            if msg_lists[0][2] == 'NEXT_POS_TRIG':
                # handle case where the trigger msg from the previous target
                # message was not read until the start of the next pos.
                # In which case, move msg to end of previous targ pos msgs
                npm = msg_lists.pop(0)
                self.target_pos_msgs[-1].append(npm)

            self.target_pos_msgs.append(msg_lists)

        for i in range(len(self.target_pos_msgs)):
            self.target_pos_msgs[i] = np.asarray(self.target_pos_msgs[i])

        return self.target_pos_msgs

    def getSampleMessageData(self):
        """Return a list of numpy ndarrays, each containing joined eye sample
        and previous / next experiment message data for the sample's time."""

        # preprocess message events
        self._processMessageEvents()

        # inline func to return sample field array based on sample namedtup
        def getSampleData(s):
            sampledata = [s.time, s.status]
            if self.sample_type == EventConstants.BINOCULAR_EYE_SAMPLE:
                sampledata.extend((s.left_gaze_x,
                                   s.left_gaze_y,
                                   s.left_pupil_measure1,
                                   s.right_gaze_x,
                                   s.right_gaze_y,
                                   s.right_pupil_measure1))
                return sampledata

            sampledata.extend((s.gaze_x,
                               s.gaze_y,
                               s.pupil_measure1))
            return sampledata

        ######

        #
        # Process Samples
        #

        current_target_pos = -1.0, -1.0
        current_targ_state = 0
        target_pos_samples = []
        for pindex, samples in enumerate(self.saved_pos_samples):
            last_msg, messages = self.target_pos_msgs[
                pindex][0], self.target_pos_msgs[pindex][1:]
            samplesforposition = []
            pos_sample_count = len(samples)
            si = 0
            for current_msg in messages:
                last_msg_time = last_msg[0]
                last_msg_type = last_msg[2]
                if last_msg_type == 'START_DRAW':
                    if not current_targ_state & self.TARGET_STATIONARY:
                        current_targ_state += self.TARGET_STATIONARY
                    current_targ_state -= current_targ_state & self.TARGET_MOVING
                    current_targ_state -= current_targ_state & self.TARGET_EXPANDING
                    current_targ_state -= current_targ_state & self.TARGET_CONTRACTING
                elif last_msg_type == 'EXPAND_SIZE':
                    if not current_targ_state & self.TARGET_EXPANDING:
                        current_targ_state += self.TARGET_EXPANDING
                    current_targ_state -= current_targ_state & self.TARGET_CONTRACTING
                elif last_msg_type == 'CONTRACT_SIZE':
                    if not current_targ_state & self.TARGET_CONTRACTING:
                        current_targ_state += self.TARGET_CONTRACTING
                    current_targ_state -= current_targ_state & self.TARGET_EXPANDING
                elif last_msg_type == 'TARGET_POS':
                    current_target_pos = float(last_msg[3]), float(last_msg[4])
                    current_targ_state -= current_targ_state & self.TARGET_MOVING
                    if not current_targ_state & self.TARGET_STATIONARY:
                        current_targ_state += self.TARGET_STATIONARY
                elif last_msg_type == 'POS_UPDATE':
                    current_target_pos = float(last_msg[3]), float(last_msg[4])
                    if not current_targ_state & self.TARGET_MOVING:
                        current_targ_state += self.TARGET_MOVING
                    current_targ_state -= current_targ_state & self.TARGET_STATIONARY
                elif last_msg_type == 'SYNCTIME':
                    if not current_targ_state & self.TARGET_STATIONARY:
                        current_targ_state += self.TARGET_STATIONARY
                    current_targ_state -= current_targ_state & self.TARGET_MOVING
                    current_targ_state -= current_targ_state & self.TARGET_EXPANDING
                    current_targ_state -= current_targ_state & self.TARGET_CONTRACTING
                    current_target_pos = float(last_msg[6]), float(last_msg[7])

                while si < pos_sample_count:
                    sample = samples[si]
                    if sample.time >= last_msg_time and sample.time < current_msg[
                            0]:
                        sarray = [pindex, last_msg_time, last_msg_type,
                                  current_msg[0], current_msg[2],
                                  current_target_pos[0], current_target_pos[1],
                                  current_targ_state]
                        sarray.extend(getSampleData(sample))
                        sndarray = np.asarray(
                            tuple(sarray), dtype=self.sample_msg_dtype)
                        samplesforposition.append(sndarray)
                        si += 1
                    elif sample.time >= current_msg[0]:
                        break
                    else:
                        si += 1
                last_msg = current_msg

            # convert any position fields to degrees if needed
            possamples = np.asanyarray(samplesforposition)
            target_pos_samples.append(possamples)

        # So we now have a list len == number target positions. Each element
        # of the list is a list of all eye sample / message data for a
        # target position. Each element of the data list for a single target
        # position is itself a list that that contains combined info about
        # an eye sample and message info valid for when the sample time was.

        return np.asanyarray(target_pos_samples)

################ Validation #######################
"""
target_graphics:
    default: &default_target_stim
        radius: 16,              # 16 pix outer radius.
        fillcolor: [.5, .5, .5]  # 75% white fill color.
        edgecolor: [-1, -1, -1]  # Fully black outer edge
        edgewidth: 3             # with a 3 pixel width.
        dotcolor: [1, -1, -1]    # Full red center dot
        dotradius: 3             # with radius of 3 pixels.
        units: pix               # Size & position units are in pix.
        colorspace: rgb          # colors are in 'rgb' space (-1.0 - 1.0) range
                                 # forevents r,g,b.
        opacity: 1.0             # The transparency of the target (0.0 - 1.0)
        contrast: 1.0            # The contrast of the target stim.

position_grids:
    default:  &default_position_grid
        shape: 3            # Defines the number of columns and rows of
                            # positions needed. If shape is an array of
                            # two elements, it defines the col,row shape
                            # for position layout. Position count will
                            # equal rows*cols. If shape is a single
                            # int, the position grid col,row shape will
                            # be shape x shape.
        posCount: None      # Defines the number of positions to
                            # without any col,row position constraint.
        leftMargin: None    # Specify the minimum valid horz position.
        rightMargin: None   # Limit horz positions to be < max horz
                            # position minus rightMargin.
        topMargin: None     # Limit vert positions to be < max vert
                            # position minus topMargin.
        bottomMargin: None  # Specify the minimum valid vert position.
        scale: 0.85         # Scale can be one or two numbers, each
                            # between 0.0 and 1.0. If a tuple is
                            # provided, it represents the horz, vert
                            # scale to be applied to window width,
                            # height. If a single number is
                            # given, the same scale will be applied to
                            # both window width and height. The scaled
                            # window size is centered on the original
                            # window size to define valid position area.
        posList: None       # Provide an existing list of (x,y)
                            # positions. If posList is provided, the
                            # shape, posCount, margin and scale arg's
                            # are ignored.
        noiseStd: None      # Add a random shift to each position based
                            # on a normal distribution with mean :  0.0
                            # and sigma equal to noiseStd. Specify
                            # value based on units being used.
        firstposindex: 4    # Specify which position in the position
                            # list should be displayed first. This
                            # position is not effected by randomization.
        repeatfirstpos: True # If the first position in the list should
                            # be provided as the last position as well,
                            # set to True. In this case, the number of
                            # positions returned will be position
                            # count + 1. False indicated the first
                            # position should not be repeated.

validation:
    target: *default_target_stim
    positions: *default_position_grid
    targ_animation:
        velocity: None             #800.0,
        expandedscale: None        #2.0,
        expansionduration: None    #0.1,
        contractionduration: None  #0.1    background: None,
    triggers:
        TimeTrigger:
            start_time: None
            delay: 1.5
        DeviceEventTrigger:
            device: keyboard
            event_type: KEYBOARD_RELEASE
            event_attribuet_conditions:
                key: space
                repeat_count: 0
    storeeventsfor: None
    accuracy_period_start: 0.550
    accuracy_period_stop: .150
    show_intro_screen: True
    intro_text: Validation procedure is now going to be performed.
    show_results_screen: True
    results_in_degrees: True
    save_figure: SESSION_RESULTS_FOLDER


"""


class ValidationProcedure(object):
    """ValidationProcedure can be used to check the accuracy of a calibrated
    eye tracking system.

    One a ValidationProcedure class instance has been created, the
    display(**kwargs) method can be called to run the full validation process.

    The validation process consists of the following stages:

    1) Display an Introduction / Instruction screen. A key press is used to
       move to the next stage.
    2) The validation target presentation sequence. Based on the Target and
       PositionGrid objects provided when the ValidationProcedure was created,
       a series of target positions are displayed. The progression from one
       target position to the next is controlled by the triggers specified.
       The target can simply jump from one position to the next, or optional
       linear motion settings can be used to have the target move across the
       screen from one point to the next. The Target graphic itself can also
       be configured to expand or contract once it has reached a location
       defined in the position grid.
    3) During stage 2), data is collected from the devices being monitored by
       iohub. Specifically eye tracker samples and experiment messages are
       collected.
    4) The data collected during the validation target sequence is used to
       calculate accuracy information for each target position presented.
       The raw data as well as the computed accuracy data is available via the
       ValidationProcedure class. Calculated measures are provided separately
       for each target position and include:

           a) An array of the samples used for the accuracy calculation. The
              samples used are selected using the following criteria:
                   i) Only samples where the target was stationary and
                      not expanding or contracting are selected.

                   ii) Samples are selected that fall between:

                          start_time_filter = last_sample_time - accuracy_period_start

                       and

                          end_time_filter = last_sample_time - accuracy_period_end

                       Therefore, the duration of the selected sample period is:

                          selection_period_dur = end_time_filter - start_time_filter

                   iii) Sample that contain missing / invalid position data
                        are then removed, providing the final set of samples
                        used for accuracy calculations. The min, max, and mean
                        values from each set of selected samples is calculated.

           b) The x and y error of each samples gaze position relative to the
              current target position. This data is in the same units as is
              used by the Target instance. Computations are done for each eye
              being recorded. The values are signed floats.

           c) The xy distance error from the from each eye's gaze position to
              the target position. This is also calculated as an average of
              both eyes when binocular data is available. The data is unsigned,
              providing the absolute distance from gaze to target positions

    5) A 2D plot is created displaying each target position and the position of
       each sample used for the accuracy calculation. The minimum, maximum, and
       average error is displayed for all target positions. A key press is used
       to remove the validation results plot, and control is returned to the
       script that started the validation display. Note that the plot is also
       saved as a png file in the same directory as the calling stript.

    See the validation.py demo in demos.coder.iohub.eyetracker for example usage.

    """

    def __init__(
            self,
            win=None,
            target=None,
            positions=None,
            target_animation_params={},
            randomize_positions=True,
            background=None,
            triggers=2.0,
            storeeventsfor=None,
            accuracy_period_start=0.350,
            accuracy_period_stop=.050,
            show_intro_screen=True,
            intro_text='Validation procedure is now going to be performed.',
            show_results_screen=True,
            results_in_degrees=False,
            save_figure=None,
            configuration=None):
        self.io = ioHubConnection.getActiveConnection()
        self.randomize_positions = randomize_positions
        self.positions = positions

        if configuration:
            cfg = configuration
            target = TargetStim(win, **cfg.get('target'))
            self.randomize_positions = cfg.get('positions').get('randomize')
            del cfg.get('positions')['randomize']
            self.positions = PositionGrid(win.size, **cfg.get('positions'))
            target_animation_params = cfg.get('targ_animation')

            triggers = cfg.get('triggers')
            trigger_list = []
            for trig_type, trig_kwargs in triggers.items():
                if trig_type == 'DeviceEventTrigger':
                    device_name = trig_kwargs.get('device')
                    device = self.io.getDevice(device_name)
                    event_type = EventConstants.getID(
                        trig_kwargs.get('event_type'))
                    det = DeviceEventTrigger(
                        device, event_type, trig_kwargs.get('event_attribute_conditions'))
                    trigger_list.append(det)
                elif trig_type == 'TimeTrigger':
                    start_time = trig_kwargs.get('start_time')
                    delay = trig_kwargs.get('delay')
                    tit = TimeTrigger(start_time, delay)
                    trigger_list.append(tit)
            triggers = trigger_list
            background = cfg.get('background')
            storeeventsfor = cfg.get('storeeventsfor')
            accuracy_period_start = cfg.get('accuracy_period_start')
            accuracy_period_stop = cfg.get('accuracy_period_stop')
            show_intro_screen = cfg.get('show_intro_screen')
            intro_text = cfg.get('intro_text')
            show_results_screen = cfg.get('show_results_screen')
            results_in_degrees = cfg.get('results_in_degrees')
            save_figure = cfg.get('save_figure')

        self.win = win
        self.display_size = win.size
        if target_animation_params is None:
            target_animation_params = {}
        self.animation_params = target_animation_params

        if self.randomize_positions:
            self.positions.randomize()
        self.accuracy_period_start = accuracy_period_start
        self.accuracy_period_stop = accuracy_period_stop
        self.show_intro_screen = show_intro_screen
        self.intro_text = intro_text
        self.show_results_screen = show_results_screen
        self.results_in_degrees = results_in_degrees
        self.pix2deg = None
        self.save_figure_path = save_figure
        if self.results_in_degrees:
            display = self.io.devices.display
            ddim = display.getPhysicalDimensions()
            dhorz, dvert = ddim['width'], ddim['height']
            self.pix2deg = VisualAngleCalc((dhorz, dvert),
                                           self.display_size,
                                           display.getDefaultEyeDistance()
                                           ).pix2deg

        self.validation_results = None
        if storeeventsfor is None:
            storeeventsfor = [self.io.devices.keyboard,
                              self.io.devices.mouse,
                              self.io.devices.tracker,
                              self.io.devices.experiment
                              ]

        # Create the TargetPosSequenceStim instance; used to control the sequential
        # presentation of the target at each of the grid positions.
        self.targetsequence = TargetPosSequenceStim(
            win,
            target=target,
            positions=self.positions,
            background=background,
            triggers=triggers,
            storeeventsfor=storeeventsfor)

        # Stim for results screen
        self.imagestim = None
        self.textstim = None

        self.use_dpi = 80

    def _waitForTrigger(self, trigger_key):
        # TODO: should process a Trigger object. Right now only kb
        # event supported.
        self.io.clearEvents('all')
        show_screen = True
        while show_screen:
            kb_events = self.io.devices.keyboard.getEvents()
            for kbe in kb_events:
                if kbe.char == trigger_key:
                    show_screen = False
                    break
            core.wait(0.2, 0.025)
        self.io.clearEvents('all')

    def display(self, **kwargs):
        """Begin the validation procedure.

        The method returns after the full
        validation process is complete, including:
            a) display of an instruction screen
            b) display of the target position sequence used for validation
               data collection.
            c) display of a validation accuracy results plot.

        """

        if self.show_intro_screen:
            # Display Validation Intro Screen
            #
            self.showIntroScreen()
            self._waitForTrigger(' ')

        # Perform Validation.....
        self.targetsequence.display(**self.animation_params)
        self.io.clearEvents('all')

        if self.show_results_screen:
            # Display Accuracy Results Plot
            if self.showResultsScreen() is not None:
                self._waitForTrigger(' ')
                return True
            return False

        return self.validation_results

    def _generateImageName(self):
        import datetime
        file_name = 'validation_' + datetime.datetime.now().strftime('%d_%m_%Y_%H_%M') + '.png'
        if self.save_figure_path:
            return normjoin(self.save_figure_path, file_name)
        rootScriptPath = os.path.dirname(sys.argv[0])
        return normjoin(rootScriptPath, file_name)

    def _createPlot(self):
        try:
            self.validation_results = None
            sample_array = self.targetsequence.getSampleMessageData()
            if self.results_in_degrees:
                for postdat in sample_array:
                    postdat['targ_pos_x'], postdat['targ_pos_y'] = self.pix2deg(
                        postdat['targ_pos_x'], postdat['targ_pos_y'])

                    postdat['left_eye_x'], postdat['left_eye_y'] = self.pix2deg(
                        postdat['left_eye_x'], postdat['left_eye_y'])

                    postdat['right_eye_x'], postdat['right_eye_y'] = self.pix2deg(
                        postdat['right_eye_x'], postdat['right_eye_y'])

            pixw, pixh = self.display_size
            # Validation Accuracy Analysis

            from matplotlib import pyplot as pl
            pl.clf()
            fig = pl.gcf()
            fig.set_size_inches(
                (pixw * .9) / self.use_dpi,
                (pixh * .8) / self.use_dpi)

            cm = pl.cm.get_cmap('RdYlBu')

            min_error = 100000.0
            max_error = 0.0
            summed_error = 0.0
            point_count = 0

            self.io.sendMessageEvent('Results', 'VALIDATION')
            results = dict(
                display_size=self.display_size,
                position_count=len(sample_array),
                positions_failed_processing=0,
                target_positions=[
                    p for p in self.targetsequence.positions],
                position_results=[])

            self.io.sendMessageEvent(
                'display_size: {0}'.format(
                    self.display_size), 'VALIDATION')
            self.io.sendMessageEvent(
                'target position_count: {0}'.format(
                    len(sample_array)), 'VALIDATION')
            self.io.sendMessageEvent('target_positions: {0}'.format(
                [p for p in self.targetsequence.positions]), 'VALIDATION')

            for pindex, samplesforpos in enumerate(sample_array):
                self.io.sendMessageEvent(
                    'Target Position Results: {0}'.format(pindex), 'VALIDATION')

                stationary_samples = samplesforpos[
                    samplesforpos['targ_state'] == self.targetsequence.TARGET_STATIONARY]

                last_stime = stationary_samples[-1]['eye_time']
                first_stime = stationary_samples[0]['eye_time']

                filter_stime = last_stime - self.accuracy_period_start
                filter_etime = last_stime - self.accuracy_period_stop

                all_samples_for_accuracy_calc = stationary_samples[
                    stationary_samples['eye_time'] >= filter_stime]
                all_samples_for_accuracy_calc = all_samples_for_accuracy_calc[
                    all_samples_for_accuracy_calc['eye_time'] < filter_etime]

                good_samples_for_accuracy_calc = all_samples_for_accuracy_calc[
                    all_samples_for_accuracy_calc['eye_status'] <= 1]

                all_samples_for_accuracy_count = all_samples_for_accuracy_calc.shape[
                    0]
                good_accuracy_sample_count = good_samples_for_accuracy_calc.shape[
                    0]
                accuracy_calc_good_sample_perc = good_accuracy_sample_count / \
                    float(all_samples_for_accuracy_count)

                # stationary_period_sample_count=stationary_samples.shape[0]
                # target_period_sample_count=samplesforpos.shape[0]
                # print 'target_period_sample_count:',target_period_sample_count
                # print 'stationary_period_sample_count:',stationary_period_sample_count
                # print 'all_samples_for_accuracy_count:',all_samples_for_accuracy_count
                # print 'good_accuracy_sample_count:',good_accuracy_sample_count
                # print
                # 'accuracy_calc_good_sample_perc:',accuracy_calc_good_sample_perc

                # Ordered dictionary of the different levels of samples
                # selected during filtering for valid samples to use in
                # accuracy calculations.
                sample_msg_data_filtering = OrderedDict(
                    # All samples from target period.
                    all_samples=samplesforpos,
                    # Sample during stationary period at
                    # end of target presentation display.
                    stationary_samples=stationary_samples,
                    # Samples that occurred within the
                    # defined time selection period.
                    time_filtered_samples=all_samples_for_accuracy_calc,
                    # Samples from the selection period that
                    # do not have missing data
                    used_samples=good_samples_for_accuracy_calc
                )

                position_results = dict(
                    pos_index=pindex,
                    sample_time_range=[
                        first_stime,
                        last_stime],
                    filter_samples_time_range=[
                        filter_stime,
                        filter_etime],
                    sample_from_filter_stages=sample_msg_data_filtering,
                    valid_filtered_sample_perc=accuracy_calc_good_sample_perc)
                self.io.sendMessageEvent('sample_time_range: {0}'.format(
                    [first_stime, last_stime]), 'VALIDATION')
                self.io.sendMessageEvent('filter_samples_time_range: {0}'.format(
                    [filter_stime, filter_etime]), 'VALIDATION')
                self.io.sendMessageEvent('valid_filtered_sample_perc: {0}'.format(
                    accuracy_calc_good_sample_perc), 'VALIDATION')

                if accuracy_calc_good_sample_perc == 0.0:
                    position_results['calculation_status'] = 'FAILED'
                    results['positions_failed_processing'] += 1
                else:
                    time = good_samples_for_accuracy_calc[:]['eye_time']

                    target_x = good_samples_for_accuracy_calc[:]['targ_pos_x']
                    target_y = good_samples_for_accuracy_calc[:]['targ_pos_y']

                    left_x = good_samples_for_accuracy_calc[:]['left_eye_x']
                    left_y = good_samples_for_accuracy_calc[:]['left_eye_y']
                    left_pupil = good_samples_for_accuracy_calc[
                        :]['left_pupil_size']
                    left_error_x = target_x - left_x
                    left_error_y = target_y - left_y
                    left_error_xy = np.hypot(left_error_x, left_error_y)

                    right_x = good_samples_for_accuracy_calc[:]['right_eye_x']
                    right_y = good_samples_for_accuracy_calc[:]['right_eye_y']
                    right_pupil = good_samples_for_accuracy_calc[
                        :]['right_pupil_size']
                    right_error_x = target_x - right_x
                    right_error_y = target_y - right_y
                    right_error_xy = np.hypot(right_error_x, right_error_y)

                    lr_x = (left_x + right_x) / 2.0
                    lr_y = (left_y + right_y) / 2.0
                    lr_error = (right_error_xy + left_error_xy) / 2.0
                    lr_error_max = lr_error.max()
                    lr_error_min = lr_error.min()
                    lr_error_mean = lr_error.mean()
                    lr_error_std = np.std(lr_error)

                    min_error = min(min_error, lr_error_min)
                    max_error = max(max_error, lr_error_max)
                    summed_error += lr_error_mean
                    point_count += 1.0

                    position_results['calculation_status'] = 'PASSED'
                    position_results['target_position'] = (
                        target_x[0], target_y[0])
                    position_results['min_error'] = lr_error_min
                    position_results['max_error'] = lr_error_max
                    position_results['mean_error'] = lr_error_mean
                    position_results['stdev_error'] = lr_error_std

                    self.io.sendMessageEvent(
                        'calculation_status: {0}'.format('PASSED'), 'VALIDATION')
                    self.io.sendMessageEvent(
                        'target_position: {0}'.format(
                            (target_x[0], target_y[0])), 'VALIDATION')
                    self.io.sendMessageEvent(
                        'min_error: {0}'.format(lr_error_min), 'VALIDATION')
                    self.io.sendMessageEvent(
                        'max_error: {0}'.format(lr_error_max), 'VALIDATION')
                    self.io.sendMessageEvent(
                        'mean_error: {0}'.format(lr_error_mean), 'VALIDATION')
                    self.io.sendMessageEvent(
                        'stdev_error: {0}'.format(lr_error_std), 'VALIDATION')
                    self.io.sendMessageEvent(
                        'Done Target Position Results : {0}'.format(pindex), 'VALIDATION')

                    normed_time = (time - time.min()) / \
                        (time.max() - time.min())

                    pl.scatter(target_x[0],
                               target_y[0],
                               s=400,
                               c=[0.75, 0.75, 0.75],
                               alpha=0.5)

                    pl.text(target_x[0], target_y[0],
                            str(pindex),
                            size=11,
                            horizontalalignment='center',
                            verticalalignment='center')

                    pl.scatter(lr_x, lr_y,
                               s=40, c=normed_time,
                               cmap=cm, alpha=0.75)

                    results['position_results'].append(position_results)

            unit_type = 'pixels'
            if self.results_in_degrees:
                ldeg, bdeg = self.pix2deg(-pixw / 2, -pixh / 2)
                rdeg, tdeg = self.pix2deg(pixw / 2, pixh / 2)
                pl.xlim(ldeg, rdeg)
                pl.ylim(bdeg, tdeg)
                unit_type = 'degrees'
            else:
                pl.xlim(-pixw / 2, pixw / 2)
                pl.ylim(-pixh / 2, pixh / 2)
            pl.xlabel('Horizontal Position (%s)' % (unit_type))
            pl.ylabel('Vertical Position (%s)' % (unit_type))

            mean_error = summed_error / point_count
            pl.title(
                'Validation Accuracy (%s)\nMin: %.2f, Max: %.2f, Mean %.2f' %
                (unit_type, min_error, max_error, mean_error))

            results['min_error'] = min_error
            results['max_error'] = max_error
            results['mean_error'] = mean_error

            self.io.sendMessageEvent(
                'Overall Results Coord Type: {0}'.format(unit_type),
                'VALIDATION')
            self.io.sendMessageEvent(
                'min_error: {0}'.format(min_error), 'VALIDATION')
            self.io.sendMessageEvent(
                'max_error: {0}'.format(max_error), 'VALIDATION')
            self.io.sendMessageEvent(
                'mean_error: {0}'.format(mean_error), 'VALIDATION')

            self.validation_results = results

            # pl.colorbar()
            fig.tight_layout()

            fig_name = self._generateImageName()
            fig.savefig(fig_name, dpi=self.use_dpi)
            self.io.sendMessageEvent(
                'Validation Plot: %s' %
                (fig_name), 'VALIDATION')
            return fig, fig_name
        except Exception:
            print('\nError While Calculating Accuracy Stats:')
            import traceback
            traceback.print_exc()
            print()

        self.io.sendMessageEvent('Validation Report Complete', 'VALIDATION')

    def getValidationResults(self):
        return self.validation_results

    def _buildResultScreen(self, replot=False):
        if replot or self.imagestim is None:
            fig, fig_image_path = self._createPlot()
            text_pos = (0, 0)
            text = 'Accuracy Calculation not Possible do to Analysis Error. Press SPACE to continue.'

            if fig:
                fig_image = Image.open(fig_image_path)

                if self.imagestim:
                    self.imagestim.setImage(fig_image)
                else:
                    self.imagestim = visual.ImageStim(self.win,
                                                      image=fig_image,
                                                      units='pix',
                                                      pos=(0.0, 0.0))

                text = 'Press SPACE to continue.'
                text_pos = (0.0, -(self.display_size[1] / 2.0) * .9)
            else:
                self.imagestim = None

            if self.textstim is None:
                self.textstim = visual.TextStim(
                    self.win,
                    text=text,
                    pos=text_pos,
                    color=(
                        0,
                        0,
                        0),
                    colorSpace='rgb255',
                    opacity=1.0,
                    contrast=1.0,
                    units='pix',
                    ori=0.0,
                    height=None,
                    antialias=True,
                    bold=False,
                    italic=False,
                    wrapWidth=self.display_size[0] * .8)
            else:
                self.textstim.setText(text)
                self.textstim.setPos(text_pos)

        elif self.imagestim:
            return True
        return False

    def showResultsScreen(self):
        self._buildResultScreen()
        if self.imagestim:
            self.imagestim.draw()
        self.textstim.draw()
        return self.win.flip()

    def showIntroScreen(self):
        text = self.intro_text + '\nPress SPACE to Start....'
        textpos = (0, 0)
        if self.textstim:
            self.textstim.setText(text)
            self.textstim.setPos(textpos)
        else:
            self.textstim = visual.TextStim(
                self.win,
                text=text,
                pos=textpos,
                height=30,
                color=(
                    0,
                    0,
                    0),
                colorSpace='rgb255',
                opacity=1.0,
                contrast=1.0,
                units='pix',
                ori=0.0,
                antialias=True,
                bold=False,
                italic=False,
                wrapWidth=self.display_size[0] * .8)

        self.textstim.draw()
        return self.win.flip()

from psychopy.iohub.util.visualangle import VisualAngleCalc



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Stimulus object for drawing arbitrary bitmap carriers with an arbitrary
second order envelope carrier and envelope can vary independently for
orientation, frequencyand phase. Also does beat stimuli. """

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2020 Open Science Tools Ltd..
# Additional code provided by Andrew Schofield
# Distributed under the terms of the GNU General Public License (GPL).

from __future__ import absolute_import, print_function

# Requires shaders if you don't have them it will just throw and error.
# Ensure setting pyglet.options['debug_gl'] to False is done prior to any
# other calls to pyglet or pyglet submodules, otherwise it may not get picked
# up by the pyglet GL engine and have no effect.
# Shaders will work but require OpenGL2.0 drivers AND PyOpenGL3.0+
import pyglet
pyglet.options['debug_gl'] = False
import ctypes
GL = pyglet.gl

import psychopy  # so we can get the __path__
from psychopy import logging

from psychopy.tools.arraytools import val2array
from psychopy.tools.attributetools import attributeSetter
from psychopy.visual.grating import GratingStim
import numpy

from . import shaders as _shaders

# we need a different shader program for this stimuli (3 textures)
carrierEnvelopeMaskFrag = '''
    uniform sampler2D carrier, envelope, mask;
    uniform float moddepth, offset, ori, add;

    void main() {
        float mid=0.5;
        vec4 carrierFrag = texture2D(carrier,gl_TexCoord[0].st);
        vec4 maskFrag = texture2D(mask,gl_TexCoord[2].st);
        vec2 tex = gl_TexCoord[1].st;

        vec2 rotated = vec2(cos(ori) * (tex.x - mid ) + sin(ori) * (tex.y - mid) + mid , cos(ori) * (tex.y - mid) - sin(ori) * (tex.x - mid) + mid );
        vec4 envFrag = texture2D( envelope,  rotated);
        gl_FragColor.a = gl_Color.a*maskFrag.a;
        gl_FragColor.rgb = ((moddepth*envFrag.rgb + offset)*carrierFrag.rgb*(gl_Color.rgb*2.0 - 1.0) + add)/2.0;
    }
    '''

# we need a different shader program for this (4 textures including the power term)
carrierEnvelopeMaskFragPow = '''
    uniform sampler2D carrier, envelope, mask, power;
    uniform float moddepth, offset, ori, add;

    void main() {
        float mid=0.5;
        vec4 carrierFrag = texture2D(carrier,gl_TexCoord[0].st);
        vec4 maskFrag = texture2D(mask,gl_TexCoord[2].st);
        vec4 powerFrag = texture2D(power,gl_TexCoord[2].st);
        vec2 tex = gl_TexCoord[1].st;

        vec2 rotated = vec2(cos(ori) * (tex.x - mid ) + sin(ori) * (tex.y - mid) + mid , cos(ori) * (tex.y - mid) - sin(ori) * (tex.x - mid) + mid );
        vec4 envFrag = texture2D( envelope,  rotated);
        gl_FragColor.a = gl_Color.a*maskFrag.a;
        gl_FragColor.rgb = (pow((moddepth*envFrag.rgb + offset), powerFrag.rgb)*carrierFrag.rgb*(gl_Color.rgb*2.0 - 1.0) + add)/2.0;
    }
    '''

class EnvelopeGrating(GratingStim):
    """Second-order envelope stimuli with 3 textures; a carrier, an envelope and a mask

    **Examples**::

    env1 = EnvelopeGrating(win,ori=0, carrier='sin', envelope='sin',
            mask = 'gauss', sf=24, envsf=4, size=1, contrast=0.5,
            moddepth=1.0, envori=0, pos=[-.5,.5],interpolate=0)
            # gives a circular patch of high frequency carrier with a
            # low frequency envelope
    env2 = EnvelopeGrating(win,ori=0, carrier=noise, envelope='sin',
            mask = None, sf=1, envsf=4, size=1, contrast=0.5,
            moddepth=0.8, envori=0, pos=[-.5,-.5],interpolate=0)
            # If noise is some numpy array containing random values gives a
            # patch of noise with a low frequency sinewave envelope
    env4 = EnvelopeGrating(win,ori=90, carrier='sin', envelope='sin',
            mask = 'gauss', sf=24, envsf=4, size=1, contrast=0.5,
            moddepth=1.0, envori=0, pos=[.5,.5], beat=True, interpolate=0)
            # Setting beat will create a second order beat stimulus which
            # critically contains no net energy at the carrier frequency
            # even though it appears to be present. In this case carrier
            # and envelope are at 90 degree to each other

    With an EnvelopeStim the carrier and envelope can have different spatial
    frequencies, phases and orientations. Its position can be shifted as a whole.

    contrast controls the contrast of the carrier and moddepth the modulation
    depth of the envelope. contrast and moddepth must work together, for moddepth=1 the max carrier
    contrast is 0.5 otherwise the displayable range will be exceeded. If moddepth < 1 higher contrasts can be accommodated.

    Opacity controls the transparency of the whole stimulus.

    Because orientation is implemented very differently for the carrier and
    envelope using this function without a broadly circular mask may produce unexpected results

    **Using EnvelopeStim with images from disk (jpg, tif, png, ...)**

    Ideally texture images to be rendered should be square with 'power-of-2'
    dimensions e.g. 16x16, 128x128. Any image that is not will be upscaled
    (with linear interpolation) to the nearest such texture by PsychoPy.
    The size of the stimulus should be specified in the normal way using
    the appropriate units (deg, pix, cm, ...). Be sure to get the aspect
    ratio the same as the image (if you don't want it stretched!).
    """

    def __init__(self,
                 win,
                 carrier="noise",
                 mask="none",
                 envelope="sin",
                 units="",
                 pos=(0.0, 0.0),
                 size=None,
                 sf=None,
                 envsf=None,
                 ori=0.0,
                 envori=0.0,
                 phase=(0.0, 0.0),
                 envphase=(0.0, 0.0),
                 beat=False,
                 texRes=128,
                 rgb=None,
                 dkl=None,
                 lms=None,
                 color=(1.0, 1.0, 1.0),
                 colorSpace='rgb',
                 contrast=0.5,  # see doc
                 moddepth=1.0,  # modulation depth for envelope
                 power=1.0, # power to raise modulator to.
                 opacity=1.0,
                 depth=0,
                 rgbPedestal=(0.0, 0.0, 0.0),
                 interpolate=False,
                 blendmode='avg',
                 name=None,
                 autoLog=None,
                 autoDraw=False,
                 maskParams=None):
        """ """  # Empty docstring. All doc is in attributes
        # what local vars are defined (these are the init params) for use by
        # __repr__
        assert win._haveShaders is True, ("Currently EnvelopeGratings needs "
                                         "your graphics card to have shaders"
                                         " and yours does not seem to.")
        self._initParams = dir()
        for unecess in ['self', 'rgb', 'dkl', 'lms']:
            self._initParams.remove(unecess)
        # initialise parent class
        GratingStim.__init__(self, win,
                             units=units, pos=pos, size=size, sf=sf,
                             ori=ori, phase=phase,
                             color=color, colorSpace=colorSpace,
                             contrast=contrast, opacity=opacity,
                             depth=depth, interpolate=interpolate,
                             name=name, autoLog=autoLog, autoDraw=autoDraw,
                             maskParams=None)
        # use shaders if available by default, this is a good thing
        self.__dict__['useShaders'] = win._haveShaders
        # UGLY HACK: Some parameters depend on each other for processing.
        # They are set "superficially" here.
        # TO DO: postpone calls to _createTexture, setColor and
        # _calcCyclesPerStim whin initiating stimulus

        #AJS
        self.__dict__['carrier'] = carrier
        self.__dict__['envelope'] = envelope
        self.__dict__['maskParams'] = maskParams

        # initialise textures and masks for stimulus
        self._carrierID = GL.GLuint()
        GL.glGenTextures(1, ctypes.byref(self._carrierID))
        self._envelopeID = GL.GLuint()
        GL.glGenTextures(1, ctypes.byref(self._envelopeID))
        self._powerID = GL.GLuint()
        GL.glGenTextures(1, ctypes.byref(self._powerID))
        self.interpolate = interpolate
        del self._texID  # created by GratingStim.__init__
        self.texRes = int(texRes)
        self.mask = mask

        self.envelope = envelope
        self.carrier = carrier
        self.envsf = val2array(envsf)
        self.envphase = val2array(envphase, False)
        self.envori = float(envori)
        self.moddepth = float(moddepth)
        self.power = float(power)
        if beat in ['True','true','Yes','yes','Y','y']:
            self.beat = True
        elif beat in ['False','false','No','no','N','n']:
            self.beat = False
        else:
            self.beat = bool(beat)
        self._needUpdate = True
        self.blendmode=blendmode
        self._shaderProgBeat = _shaders.compileProgram(
            _shaders.vertSimple, carrierEnvelopeMaskFrag)
        self._shaderProgPow = _shaders.compileProgram(
            _shaders.vertSimple, carrierEnvelopeMaskFragPow)

        self.local = numpy.ones((texRes, texRes), dtype=numpy.ubyte)
        self.local_p = self.local.ctypes
        # fix scaling to window coords
        self._calcEnvCyclesPerStim()
        
        del self.__dict__['tex']

    @attributeSetter
    def envsf(self, value):
        """Spatial frequency of the envelope texture

        Should be a :ref:`x,y-pair <attrib-xy>` or
        :ref:`scalar <attrib-scalar>` or None.
        If `units` == 'deg' or 'cm' units are in cycles per deg or cm
        as appropriate.
        If `units` == 'norm' then sf units are in cycles per stimulus
        (and so envsf scales with stimulus size).
        If texture is an image loaded from a file then envsf=None defaults
        to 1/stimSize to give one cycle of the image.
        Note sf is inhertited from GratingStim and controls the spatial frequency of the carrier
        """

        # Recode phase to numpy array
        if value is None:
            # set the sf to default e.g. 1./size of the loaded image etc

            if (self.units in ['pix', 'pixels'] or
                    self._origSize is not None and
                    self.units in ['deg', 'cm']):
                value = 1.0/self.size  # default to one cycle
            else:
                value = numpy.array([1.0, 1.0])
        else:
            value = val2array(value)

        # Set value and update stuff
        self.__dict__['envsf'] = value
        self._calcEnvCyclesPerStim()
        self._needUpdate = True

    @attributeSetter
    def envphase(self, value):
        """Phase of the modulation in each dimension of the envelope texture.

        Should be an :ref:`x,y-pair <attrib-xy>` or
        :ref:`scalar <attrib-scalar>`

        **NB** phase has modulus 1 (rather than 360 or 2*pi)
        This is a little unconventional but has the nice effect
        that setting phase=t*n drifts a stimulus at n Hz
        Note phase in inherited from GratingStim and controls the phase of the carrier
        """
        # Recode phase to numpy array
        value = val2array(value)
        self.__dict__['envphase'] = value
        self._needUpdate = True

    @attributeSetter
    def moddepth(self, value):
        """Modulation depth or 'contrast' for the envelope component (0.0 - 1.0)

        Sets the range of variation in carrier contrast.
        MD=(Cmax-Cmin)/(Cma+Cmin)
        """
        self.__dict__['moddepth'] = value
        self._needUpdate = True

    @attributeSetter
    def envori(self, value):
        """The orientation for the envelope texture (in degrees).

        Should be a single value (scalar). Operations are supported.
        Orientation convention is like a clock: 0 is vertical, and positive values rotate clockwise. Beyond 360 and below zero values wrap appropriately.
        Note ori is inhertied from gratingStim and controls only the orientation of the carrier
        """
        self.__dict__['envori'] = value
        self._needUpdate = True

    @attributeSetter
    def beat(self, value):
        """Beat (True) stimulus or full moduaiton (False)

        'The differences is that the spatial frequency components of the carrier are absent in a beat
        (although the carrier will still be a visible component of the overall image)
        whereas they are present in a full modulation. Beats will always appear to have a 100% modulation
        depth and if sinusoidal the modulation will appear to be twice the requested spatial frequency.
        The modulation depth of fully modulated stimuli can be varied and they appear at their true frequency.
        Both beats and full modulations appear in the literature and have specific uses.
        """
        if value in ['True','true','Yes','yes','Y','y']:
            self.__dict__['beat'] =True
        elif value in ['False','false','No','no','N','n']:
            self.__dict__['beat'] =False
        else:
            self.__dict__['beat'] =bool(value)
        self._needUpdate = True


    #@attributeSetter
    #def blendmode(self, value):
    #    """The OpenGL mode in which the stimulus is draw
    #
    #    Can the 'avg' or 'add'. Average (avg) places the new stimulus over the old one
    #    with a transparency given by its opacity. Opaque stimuli will hide other stimuli
    #    transparent stimuli won't. Add performs the arithmetic sum of the new stimulus and the ones
    #    already present.
    #
    #    """
    #    self.__dict__['blendmode'] = value
    #   self._needUpdate = True

    @attributeSetter
    def carrier(self, value):
        """Texture to use in the stimulus as a carrier, typically noise array.

        This can be one of various options:
            + the name of an image file (most formats supported)
            + a numpy array (1xN or NxN) ranging -1:1

        If specifying your own texture using an image or numpy array
        you should ensure that the image has square power-of-two
        dimesnions (e.g. 256 x 256). If not then PsychoPy will upsample
        your stimulus to the next larger power of two.
        """
        self._createTexture(value, id=self._carrierID, pixFormat=GL.GL_RGB,
                            stim=self, res=self.texRes)
        # if user requested size=None then update the size for new stim here
        if hasattr(self, '_requestedSize') and self._requestedSize is None:
            self.size = None  # Reset size do default
        self.__dict__['carrier'] = value
        self._needTextureUpdate = False

    @attributeSetter
    def envelope(self, value):
        """Texture to use on the stimulus as a envelope, typically a grating.

        This can be one of various options:
            + **'sin'**,'sqr', 'saw', 'tri', None (resets to default)
            + the name of an image file (most formats supported)
            + a numpy array (1xN or NxN) ranging -1:1

        If specifying your own texture using an image or numpy array
        you should ensure that the image has square power-of-two dimesnions
        (e.g. 256 x 256). If not then PsychoPy will upsample your stimulus
        to the next larger power of two.
        """
        if self.useShaders == True:
            self._createTexture(value, id=self._envelopeID,
                                pixFormat=GL.GL_RGB, stim=self,
                                res=self.texRes)
        else:
            self._createTexture(value, id=self._envelopeID,
                                pixFormat=GL.GL_ALPHA, stim=self,
                                res=self.texRes)

        # if user requested size=None then update the size for new stim here
        if hasattr(self, '_requestedSize') and self._requestedSize is None:
            self.size = None  # Reset size do default
        self.__dict__['envelope'] = value
        self._needTextureUpdate = False

    @attributeSetter
    def power(self, value):
        """Will raise the modulator to the given power 
           using the equation S=C*(1+M)^power where C is the carrier and M
           is the modulator. Only works with AM envelopes (hence +1) in 
           equation. Power is ignored if a beat is requested.
           This is used to obtain the square root of the modulator (power = 0.5)
           which is useful if combining two envelope gratings with different carriers
           and a 180 degree phase shift as the resulting combined signal will not 
           have any reduction in local contrast at any point in the image.
           This is similar - but not identical to the method used by 
           Landy and Oruc, Vis Res 2002.
           Note overall contrast (apparent carrier contrast) will be altered.
        """
        if self.useShaders == True:
            self._createTexture(numpy.ones((self.texRes,self.texRes))*value, id=self._powerID,
                                pixFormat=GL.GL_RGB, stim=self,
                                res=self.texRes)

        # if user requested size=None then update the size for new stim here
        if hasattr(self, '_requestedSize') and self._requestedSize is None:
            self.size = None  # Reset size do default
        self.__dict__['power'] = value
        self._needTextureUpdate = False

    @attributeSetter
    def texRes(self, value):
        """Power-of-two int. Sets the resolution of the mask and texture.
        texRes is overridden if an array or image is provided as mask.

        :ref:`Operations <attrib-operations>` supported.
        """
        self.__dict__['texRes'] = value

        # ... now rebuild textures (call attributeSetters without logging).
        if hasattr(self, 'carrier'):
            self._set('carrier', self.carrier, log=False)
        if hasattr(self, 'envelope'):
            self._set('envelope', self.envelope, log=False)
        #if hasattr(self, 'power'):
        #    self._set('power', self.power, log=False)
        if hasattr(self, 'mask'):
            self._set('mask', self.mask, log=False)

    def setTexRes(self, value, log=None):
        """DEPRECATED. Use 'stim.parameter = value' syntax instead
        """
        self._set('texRes', value, log=log)

    def setEnvori(self, value, log=None):
        """DEPRECATED. Use 'stim.parameter = value' syntax instead
        """
        self._set('envori', value, log=log)

    def setEnvsf(self, value, log=None):
        """DEPRECATED. Use 'stim.parameter = value' syntax instead
        """
        self._set('envsf', value, log=log)

    def setEnvphase(self, value, log=None):
        """DEPRECATED. Use 'stim.parameter = value' syntax instead
        """
        self._set('envphase', value, log=log)

    def setModdepth(self, value, log=None):
        """DEPRECATED. Use 'stim.parameter = value' syntax instead
        """
        self._set('moddepth', value, log=log)
        
    def setPower(self, value, log=None):
        """DEPRECATED. Use 'stim.parameter = value' syntax instead
        """
        self.power = numpy.ones(self.texRes,self.texRes)*value 

    def setBeat(self, value, log=None):
        """DEPRECATED. Use 'stim.parameter = value' syntax instead
        """
        self._set('beat', value, log=log)

    def setCarrier(self, value, log=None):
        """DEPRECATED. Use 'stim.parameter = value' syntax instead
        """
        self.carrier = value

    def setEnvelope(self, value, log=None):
        """DEPRECATED. Use 'stim.parameter = value' syntax instead
        """
        self.envelope = value

    #def setBlendmode(self, value, log=None):
    #    """DEPRECATED. Use 'stim.parameter = value' syntax instead
    #    """
    #    self._set('blendmode', value, log=log)

    #def draw(self, win=None):
    #    """Draw the stimulus in its relevant window. You must call
    #    this method after every MyWin.flip() if you want the
    #    stimulus to appear on that frame and then update the screen
    #    again.
    #    """
    #    if win is None:
    #        win = self.win

    #    saveBlendMode=win.blendMode
    #    win.blendMode=self.blendmode
     #   self._selectWindow(win)
    #
    #    super(EnvelopeGrating, self).draw(win)

    #    win.blendMode=saveBlendMode

    def _updateListShaders(self):
        """
        The user shouldn't need this method since it gets called
        after every call to .set() Basically it updates the OpenGL
        representation of your stimulus if some parameter of the
        stimulus changes. Call it if you change a property manually
        rather than using the .set() command
        """

        # make some corrections for the envelope:: This could be done whenever
        # the envelope variables are set using some internal variables,
        # putting it here is safer but sometimes slower

        # correct envelope orientation to adjust for link with carrier
        # orientation, to make it clockwise handed and convert to radians
        envrad = (self.ori - self.envori) * numpy.pi / 180.0

        # adjust envelope phases so that any envelope drift points
        # in the same direction as the envelope.
        rph1 = (numpy.cos(envrad) *
                self.envphase[0] + numpy.sin(envrad) * self.envphase[1])
        rph2 = (-numpy.cos(envrad) *
                self.envphase[1] + numpy.sin(envrad) * self.envphase[0])

        # we need a corrective offset when the blendmode is set to average
        if self.blendmode=='avg':
            addvalue = 1.0
        else:
            addvalue = 0.0
            
        # setup the shaderprogram
        if self.beat:
            self._shaderProg = self._shaderProgBeat
        else:
            self._shaderProg = self._shaderProgPow
        self._needUpdate = False
        GL.glNewList(self._listID, GL.GL_COMPILE)
        GL.glUseProgram(self._shaderProg)

        # set the carrier to be texture unit 0
        GL.glUniform1i(GL.glGetUniformLocation(self._shaderProg, b"carrier"),
                       0)
        # set the envelope to be texture unit 1
        GL.glUniform1i(GL.glGetUniformLocation(
            self._shaderProg, b"envelope"), 1)
        GL.glUniform1i(GL.glGetUniformLocation(
            self._shaderProg, b"mask"), 2)  # mask is texture unit 2
        GL.glUniform1i(GL.glGetUniformLocation(
            self._shaderProg, b"power"), 3)  # power is texture unit 3
        GL.glUniform1f(GL.glGetUniformLocation(
            self._shaderProg, b"moddepth"), self.moddepth)
        GL.glUniform1f(GL.glGetUniformLocation(
            self._shaderProg, b"ori"), envrad)
        # CM envelopes use (modedepth*envelope+1.0)*carrier. If beat is True
        # this becomes (moddepth*envelope)*carrier thus maing a second order
        # 'beat' pattern.
        if self.beat:
            GL.glUniform1f(GL.glGetUniformLocation(
                self._shaderProg, b"offset"),0.0)
        else:
            GL.glUniform1f(GL.glGetUniformLocation(
                self._shaderProg, b"offset"), 1.0)
        GL.glUniform1f(GL.glGetUniformLocation(
            self._shaderProg, b"add"), addvalue)

        # mask
        GL.glActiveTexture(GL.GL_TEXTURE2)
        GL.glBindTexture(GL.GL_TEXTURE_2D, self._maskID)
        GL.glEnable(GL.GL_TEXTURE_2D)  # implicitly disables 1D
        # envelope
        GL.glActiveTexture(GL.GL_TEXTURE1)
        GL.glBindTexture(GL.GL_TEXTURE_2D, self._envelopeID)
        GL.glEnable(GL.GL_TEXTURE_2D)
        #power
        GL.glActiveTexture(GL.GL_TEXTURE3)
        GL.glBindTexture(GL.GL_TEXTURE_2D, self._powerID)
        GL.glEnable(GL.GL_TEXTURE_2D)
        # carrier
        GL.glActiveTexture(GL.GL_TEXTURE0)
        GL.glBindTexture(GL.GL_TEXTURE_2D, self._carrierID)
        GL.glEnable(GL.GL_TEXTURE_2D)

        Lcar = (-self._cycles[0]/2) - self.phase[0] + 0.5
        Rcar = (+self._cycles[0]/2) - self.phase[0] + 0.5
        Tcar = (+self._cycles[1]/2) - self.phase[1] + 0.5
        Bcar = (-self._cycles[1]/2) - self.phase[1] + 0.5

        Lenv = (-self._envcycles[0]/2) - rph1 + 0.5
        Renv = (+self._envcycles[0]/2) - rph1 + 0.5
        Tenv = (+self._envcycles[1]/2) - rph2 + 0.5
        Benv = (-self._envcycles[1]/2) - rph2 + 0.5
        Lmask = Bmask = 0.0
        Tmask = Rmask = 1.0  # mask

        # access just once because it's slower than basic property
        vertsPix = self.verticesPix
        GL.glBegin(GL.GL_QUADS)                  # draw a 4 sided polygon
        # right bottom
        GL.glMultiTexCoord2f(GL.GL_TEXTURE0, Rcar, Bcar)
        GL.glMultiTexCoord2f(GL.GL_TEXTURE1, Renv, Benv)
        GL.glMultiTexCoord2f(GL.GL_TEXTURE2, Rmask, Bmask)
        GL.glVertex2f(vertsPix[0, 0], vertsPix[0, 1])
        # left bottom
        GL.glMultiTexCoord2f(GL.GL_TEXTURE0, Lcar, Bcar)
        GL.glMultiTexCoord2f(GL.GL_TEXTURE1, Lenv, Benv)
        GL.glMultiTexCoord2f(GL.GL_TEXTURE2, Lmask, Bmask)
        GL.glVertex2f(vertsPix[1, 0], vertsPix[1, 1])
        # left top
        GL.glMultiTexCoord2f(GL.GL_TEXTURE0, Lcar, Tcar)
        GL.glMultiTexCoord2f(GL.GL_TEXTURE1, Lenv, Tenv)
        GL.glMultiTexCoord2f(GL.GL_TEXTURE2, Lmask, Tmask)
        GL.glVertex2f(vertsPix[2, 0], vertsPix[2, 1])
        # right top
        GL.glMultiTexCoord2f(GL.GL_TEXTURE0, Rcar, Tcar)
        GL.glMultiTexCoord2f(GL.GL_TEXTURE1, Renv, Tenv)
        GL.glMultiTexCoord2f(GL.GL_TEXTURE2, Rmask, Tmask)
        GL.glVertex2f(vertsPix[3, 0], vertsPix[3, 1])
        GL.glEnd()

        # unbind the textures
        #power
        GL.glActiveTexture(GL.GL_TEXTURE3)
        GL.glBindTexture(GL.GL_TEXTURE_2D, 0)
        GL.glDisable(GL.GL_TEXTURE_2D)
        # mask
        GL.glActiveTexture(GL.GL_TEXTURE2)
        GL.glBindTexture(GL.GL_TEXTURE_2D, 0)
        GL.glDisable(GL.GL_TEXTURE_2D)
        # envelope
        GL.glActiveTexture(GL.GL_TEXTURE1)
        GL.glBindTexture(GL.GL_TEXTURE_2D, 0)
        GL.glDisable(GL.GL_TEXTURE_2D)
        # main carrier
        GL.glActiveTexture(GL.GL_TEXTURE0)
        GL.glBindTexture(GL.GL_TEXTURE_2D, 0)
        GL.glDisable(GL.GL_TEXTURE_2D)

        GL.glUseProgram(0)

        GL.glEndList()

    # for the sake of older graphics cards------------------------------------
    def _updateListNoShaders(self):
        """EnvelopeGratings require shaders so this function should never be reached.
        It currently combines the carrier and envelope as if they
        add, so is plain wrong. Therefore there is an assertion in the
        init function which will throw an error if the window object does
        not have shaders. If someone without shaders wishes to do second-order
        gratings they need to provide a new solution.
        """
        #The user shouldn't need this method since it gets called
        #after every call to .set() Basically it updates the OpenGL
        #representation of your stimulus if some parameter of the
        #stimulus changes. Call it if you change a property manually
        #rather than using the .set() command
        #"""
        self._needUpdate = False

        # glColor can interfere with multitextures
        GL.glColor4f(1.0, 1.0, 1.0, 1.0)
        # mask
        GL.glActiveTexture(GL.GL_TEXTURE2)
        GL.glEnable(GL.GL_TEXTURE_2D)  # implicitly disables 1D
        GL.glBindTexture(GL.GL_TEXTURE_2D, self._maskID)
        # envelope (eg a grating but can be anything)
        GL.glActiveTexture(GL.GL_TEXTURE1)
        GL.glEnable(GL.GL_TEXTURE_2D)  # implicitly disables 1D
        GL.glBindTexture(GL.GL_TEXTURE_2D, self._envelopeID)
        # carrier (eg noise or textuture)
        GL.glActiveTexture(GL.GL_TEXTURE0)
        GL.glEnable(GL.GL_TEXTURE_2D)
        GL.glBindTexture(GL.GL_TEXTURE_2D, self._carrierID)

        # depth = self.depth

        Lcar = (-self._cycles[0]/2) - self.phase[0] + 0.5
        Rcar = (+self._cycles[0]/2) - self.phase[0] + 0.5
        Tcar = (+self._cycles[1]/2) - self.phase[1] + 0.5
        Bcar = (-self._cycles[1]/2) - self.phase[1] + 0.5

        Lenv = (-self._envcycles[0]/2) - self.envphase[0] + 0.5
        Renv = (+self._envcycles[0]/2) - self.envphase[0] + 0.5
        Tenv = (+self._envcycles[1]/2) - self.envphase[1] + 0.5
        Benv = (-self._envcycles[1]/2) - self.envphase[1] + 0.5

        Lmask = Bmask = 0.0
        Tmask = Rmask = 1.0  # mask

        # access just once because it's slower than basic property
        vertsPix = self.verticesPix
        # draw a 4 sided polygon
        GL.glBegin(GL.GL_QUADS)
        # right bottom
        GL.glMultiTexCoord2f(GL.GL_TEXTURE0, Rcar, Bcar)
        GL.glMultiTexCoord2f(GL.GL_TEXTURE1, Renv, Benv)
        GL.glMultiTexCoord2f(GL.GL_TEXTURE2, Rmask, Bmask)
        GL.glVertex2f(vertsPix[0, 0], vertsPix[0, 1])
        # left bottom
        GL.glMultiTexCoord2f(GL.GL_TEXTURE0, Lcar, Bcar)
        GL.glMultiTexCoord2f(GL.GL_TEXTURE1, Lenv, Benv)
        GL.glMultiTexCoord2f(GL.GL_TEXTURE2, Lmask, Bmask)
        GL.glVertex2f(vertsPix[1, 0], vertsPix[1, 1])
        # left top
        GL.glMultiTexCoord2f(GL.GL_TEXTURE0, Lcar, Tcar)
        GL.glMultiTexCoord2f(GL.GL_TEXTURE1, Lenv, Tenv)
        GL.glMultiTexCoord2f(GL.GL_TEXTURE2, Lmask, Tmask)
        GL.glVertex2f(vertsPix[2, 0], vertsPix[2, 1])
        # right top
        GL.glMultiTexCoord2f(GL.GL_TEXTURE0, Rcar, Tcar)
        GL.glMultiTexCoord2f(GL.GL_TEXTURE1, Renv, Tenv)
        GL.glMultiTexCoord2f(GL.GL_TEXTURE2, Rmask, Tmask)
        GL.glVertex2f(vertsPix[3, 0], vertsPix[3, 1])
        GL.glEnd()

        # disable mask
        GL.glActiveTexture(GL.GL_TEXTURE2)
        GL.glDisable(GL.GL_TEXTURE_2D)
        GL.glBindTexture(GL.GL_TEXTURE_2D, 0)

        # disable mask
        GL.glActiveTexture(GL.GL_TEXTURE1)
        GL.glDisable(GL.GL_TEXTURE_2D)
        GL.glBindTexture(GL.GL_TEXTURE_2D, 0)

        # main texture
        GL.glActiveTexture(GL.GL_TEXTURE0)
        GL.glDisable(GL.GL_TEXTURE_2D)
        GL.glBindTexture(GL.GL_TEXTURE_2D, 0)

        # we're done!
        GL.glEndList()

    def clearTextures(self):
        """This will be used by the __del__ method of EnvelopeGrating
        """
        GL.glDeleteTextures(1, self._carrierID)
        GL.glDeleteTextures(1, self._envelopeID)
        GL.glDeleteTextures(1, self._maskID)

    def _calcEnvCyclesPerStim(self):
        """The user should never need to call this function directly as it is
        called whenever there is a need to recalcuate the spatial
        frequency of the envelope.
        """
        if self.units in ('norm', 'height'):
            # self._cycles = self.sf  # this is the only form of sf that is
            # not size dependent
            self._envcycles = self.envsf
        else:
            # self._cycles = self.sf * self.size
            self._envcycles = self.envsf * self.size



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
gc_cursor_demo/run.py

Demonstrates the ioHub Common EyeTracking Interface by displaying a gaze cursor
at the currently reported gaze position on an image background.
All currently supported Eye Tracker Implementations are supported,
with the Eye Tracker Technology chosen at the start of the demo via a
drop down list. Exact same demo script is used regardless of the
Eye Tracker hardware used.

Initial Version: May 6th, 2013, Sol Simpson
Updated: July 30th, Sol
"""

from __future__ import absolute_import, division, print_function

from psychopy import visual
from psychopy.data import TrialHandler,importConditions
from psychopy.iohub import ioHubExperimentRuntime, module_directory
from psychopy.iohub.util import getCurrentDateTimeString
import os

class ExperimentRuntime(ioHubExperimentRuntime):
    """
    Create an experiment using psychopy and the ioHub framework by extending
    the ioHubExperimentRuntime class and implementing the run() method.
    """
    def run(self,*args):
        """
        The run method contains your experiment logic. In this example we:

        1) Load an xlsx file containing the trial conditions for use
           during the experiment. All DV's and IV's to be used or updated
           for each trial must be specified as columns in the xlsx file.
        2) Inform the ioDataStore of the trial conditions to be used, resulting in the
           creation of an experiment specific results table, with a field for each
           DV and IV defined in the xls file.
        3) Run the eye tracking device's runSetupProcedure(), which allows
           the calibration, validation, etc. of the eye tracking system being used.
        4) Create the experiment runtime graphics, including creating a cache of
           images to be displayed for each trial of the experiment.
        5) Run the experimental block of trials of the demo. Each trial sequence
           consists of:
               a) The participant pressing the SPACE key to start the trial.
               b) Randomly displaying one of the background images for a trial.
               c) Starting recording of data from the eye tracker.
               d) Displaying a gaze contingent dot located at the gaze position reported by the eye tracker.
               e) Ending each trial by pressing the SPACE key.
               f) Sending any condition variable value changes for that trial
                  to the ioDataStore for easy future selection of device events
                  recorded during the trial or for specific condition variable values.
               g) Stopping of event recording on the eye tracker device.
        """

        exp_conditions=importConditions('trial_conditions.xlsx')
        trials = TrialHandler(exp_conditions,1)

        # Inform the ioDataStore that the experiment is using a
        # TrialHandler. The ioDataStore will create a table
        # which can be used to record the actual trial variable values (DV or IV)
        # in the order run / collected.
        #
        self.hub.createTrialHandlerRecordTable(trials)

        # Let's make some short-cuts to the devices we will be using
        # in this demo.
        try:
            tracker=self.hub.devices.tracker
        except Exception:
            # No eye tracker config found in iohub_config.yaml
            from psychopy.iohub.util import MessageDialog
            md = MessageDialog(title="No Eye Tracker Configuration Found",
                               msg="Update the iohub_config.yaml file by "
                               "uncommenting\nthe appropriate eye tracker "
                               "config lines.\n\nPress OK to exit demo.",
                               showButtons=MessageDialog.OK_BUTTON,
                               dialogType=MessageDialog.ERROR_DIALOG,
                               allowCancel=False,
                               display_index=0)
            md.show()
            return 1

        display=self.hub.devices.display
        kb=self.hub.devices.keyboard
        mouse=self.hub.devices.mouse

        # Start by running the eye tracker default setup procedure.
        # The details of the setup procedure (calibration, validation, etc)
        # are unique to each implementation of the Common Eye Tracker Interface.
        # All have the common end goal of calibrating the eye tracking system
        # prior to data collection.
        # Please see the eye tracker interface implementation details for the
        # hardware being used at:
        # http://www.isolver-solutions.com/iohubdocs/iohub/api_and_manual/device_details/eyetracker.html#eye-tracking-hardware-implementations
        #
        tracker.runSetupProcedure()

        # Create a psychopy window for the experiment graphics,
        # ioHub supports the use of one full screen window during
        # the experiment runtime. (If you are using a window at all).
        #
        res=display.getPixelResolution() # Current pixel resolution of the Display to be used
        coord_type=display.getCoordinateType()
        window=visual.Window(res,monitor=display.getPsychopyMonitorName(), # name of the PsychoPy Monitor Config file if used.
                                    units=coord_type, # coordinate space to use.
                                    fullscr=True, # We need full screen mode.
                                    allowGUI=False, # We want it to be borderless
                                    screen= display.getIndex() # The display index to use, assuming a multi display setup.
                                    )

        # Create a dict of image stim for trials and a gaze blob to show the
        # reported gaze position with.
        #
        image_cache=dict()
        image_names=['canal.jpg','fall.jpg','party.jpg','swimming.jpg','lake.jpg']
        for iname in image_names:
            image_cache[iname]=visual.ImageStim(window, image=os.path.join('./images/',iname),
                        name=iname,units=coord_type)

        # Create a circle to use for the Gaze Cursor. Current units assume pix.
        #
        gaze_dot =visual.GratingStim(window,tex=None, mask="gauss",
                                     pos=(0,0 ),size=(66,66),color='green',
                                                        units=coord_type)

        # Create a Text Stim for use on /instruction/ type screens.
        # Current units assume pix.
        instructions_text_stim = visual.TextStim(window, text='', pos=[0,0],
                                    height=24,
                                    color=[-1,-1,-1], colorSpace='rgb',
                                    wrapWidth=window.size[0]*.9)


        # Update Instruction Text and display on screen.
        # Send Message to ioHub DataStore with Exp. Start Screen display time.
        #
        instuction_text="Press Any Key to Start Experiment."
        instructions_text_stim.setText(instuction_text)
        instructions_text_stim.draw()
        flip_time=window.flip()
        self.hub.sendMessageEvent(text="EXPERIMENT_START",sec_time=flip_time)

        # Wait until a key event occurs after the instructions are displayed
        self.hub.clearEvents('all')
        kb.waitForPresses()

        # Send some information to the ioDataStore as experiment messages,
        # including the experiment and session id's, the calculated pixels per
        # degree, display resolution, etc.
        #
        self.hub.sendMessageEvent(text="IO_HUB EXPERIMENT_INFO START")
        self.hub.sendMessageEvent(text="ioHub Experiment started {0}".format(getCurrentDateTimeString()))
        self.hub.sendMessageEvent(text="Experiment ID: {0}, Session ID: {1}".format(self.hub.experimentID,self.hub.experimentSessionID))
        self.hub.sendMessageEvent(text="Stimulus Screen ID: {0}, Size (pixels): {1}, CoordType: {2}".format(display.getIndex(),display.getPixelResolution(),display.getCoordinateType()))
        self.hub.sendMessageEvent(text="Calculated Pixels Per Degree: {0} x, {1} y".format(*display.getPixelsPerDegree()))
        self.hub.sendMessageEvent(text="IO_HUB EXPERIMENT_INFO END")

        self.hub.clearEvents('all')

        # For each trial in the set of trials within the current block.
        #
        t=0
        for trial in trials:
            # Update the instruction screen text to indicate
            # a trial is about to start.
            #
            instuction_text="Press Space Key To Start Trial %d"%t
            instructions_text_stim.setText(instuction_text)
            instructions_text_stim.draw()
            flip_time=window.flip()
            self.hub.sendMessageEvent(text="EXPERIMENT_START",sec_time=flip_time)


            # Wait until a space key press event occurs after the
            # start trial instuctions have been displayed.
            #
            self.hub.clearEvents('all')
            kb.waitForPresses(keys=[' ',])


            # Space Key has been pressed, start the trial.
            # Set the current session and trial id values to be saved
            # in the ioDataStore for the upcoming trial.
            #

            trial['session_id']=self.hub.getSessionID()
            trial['trial_id']=t+1

            # Send a msg to the ioHub indicating that the trial started, and the time of
            # the first retrace displaying the trial stm.
            #
            self.hub.sendMessageEvent(text="TRIAL_START",sec_time=flip_time)


            # Start Recording Eye Data
            #
            tracker.setRecordingState(True)

            # Get the image stim for this trial.
            #
            imageStim=image_cache[trial['IMAGE_NAME']]
            imageStim.draw()
            flip_time=window.flip()
            # Clear all the events received prior to the trial start.
            #
            self.hub.clearEvents('all')
            # Send a msg to the ioHub indicating that the trial started,
            # and the time of the first retrace displaying the trial stim.
            #
            self.hub.sendMessageEvent(text="TRIAL_START",sec_time=flip_time)
            # Set the value of the trial start variable for this trial
            #
            trial['TRIAL_START']=flip_time

            # Loop until we get a keyboard event
            #
            run_trial=True
            while run_trial is True:
                # Get the latest gaze position in display coord space..
                #
                gpos=tracker.getPosition()
                if type(gpos) in [tuple,list]:
                    # If we have a gaze position from the tracker,
                    # redraw the background image and then the
                    # gaze_cursor at the current eye position.
                    #
                    gaze_dot.setPos([gpos[0],gpos[1]])
                    imageStim.draw()
                    gaze_dot.draw()
                else:
                    # Otherwise just draw the background image.
                    # This will remove the gaze cursor from the screen
                    # when the eye tracker is not successfully
                    # tracking eye position.
                    #
                    imageStim.draw()

                # Flip video buffers, displaying the stim we just
                # updated.
                #
                flip_time=window.flip()

                # Send an experiment message to the ioDataStore
                # indicating the time the image was drawn and
                # current position of gaze spot.
                #
                if type(gpos) in [tuple,list]:
                    self.hub.sendMessageEvent("IMAGE_UPDATE %s %.3f %.3f"%(
                                                trial['IMAGE_NAME'],gpos[0],gpos[1]),
                                                sec_time=flip_time)
                else:
                    self.hub.sendMessageEvent("IMAGE_UPDATE %s [NO GAZE]"%(
                                                trial['IMAGE_NAME']),
                                                sec_time=flip_time)

                # Check any new keyboard press events by a space key.
                # If one is found, set the trial end variable and break.
                # from the loop
                if kb.getPresses(keys=[' ',]):
                    run_trial=False
                    break

            # The trial has ended, so update the trial end time condition value,
            # and send a message to the ioDataStore with the trial end time.
            #
            flip_time=window.flip()
            trial['TRIAL_END']=flip_time
            self.hub.sendMessageEvent(text="TRIAL_END %d"%t,sec_time=flip_time)

            # Stop recording eye data.
            # In this example, we have no use for any eye data
            # between trials, so why save it.
            #
            tracker.setRecordingState(False)

            # Save the experiment condition variable values for this
            # trial to the ioDataStore.
            #
            self.hub.addTrialHandlerRecord(trial)

            # Clear all event buffers
            #
            self.hub.clearEvents('all')
            t+=1

        # All trials have been run, so end the experiment.
        #

        flip_time=window.flip()
        self.hub.sendMessageEvent(text='EXPERIMENT_COMPLETE',sec_time=flip_time)

        # Disconnect the eye tracking device.
        #
        tracker.setConnectionState(False)

        # The experiment is done, all trials have been run.
        # Clear the screen and show an 'experiment  done' message using the
        # instructionScreen text.
        #
        instuction_text="Press Any Key to Exit Demo"
        instructions_text_stim.setText(instuction_text)
        instructions_text_stim.draw()
        flip_time=window.flip()
        self.hub.sendMessageEvent(text="SHOW_DONE_TEXT",sec_time=flip_time)
        self.hub.clearEvents('all')
        # wait until any key is pressed
        kb.waitForPresses()
    ### End of experiment logic


####### Launch the Experiment #######

runtime=ExperimentRuntime(module_directory(ExperimentRuntime.run), "experiment_config.yaml")
runtime.start()




# coding=utf-8
# Part of the psychopy.iohub library.
# Copyright (C) 2012-2016 iSolver Software Solutions
# Distributed under the terms of the GNU General Public License (GPL).
from __future__ import division, absolute_import, print_function

import sys
import traceback


def print2err(*args):
    """
    Note: As of atleast Jan-2020, this function seems to 
    cause iohub to fail to start if the script is started from Coder. 
    try: except: atleast stops whatever is crashing, 
    (Appears to be use of sys.stderr.write) but prints() do not appear in 
    Coder Console. Not sure how to get iohub process prints to 
    appear in Builder Console...??? Issue is specific to running script from Coder.
    
    Using the standard python print() function from the iohub server process
    will not print anything to the psychopy process stdout. Use print2err
    for this purpose. Each element of *args is unicode formatted and then
    written to sys.stderr.
    
    :param args: 0 to N objects of any type.
    """
    try:
        for a in args:
            sys.stderr.write("{0}".format(a))
        sys.stderr.write("\n")
        sys.stderr.flush()
    except:
        for a in args:
            print("{0}".format(a))
        print()
        	
def printExceptionDetailsToStdErr():
    """
    Print the last raised exception in the iohub (well, calling) process
    to the psychopy process stderr.
    """
    try:
        traceback.print_exc(file=sys.stderr)
        sys.stderr.flush()
    except:
        traceback.print_exc()

class ioHubError(Exception):
    #TODO: Fix the way exceptions raised in the iohub process are handled
    #      and reported to the psychopy process.
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args)
        self.args = args
        self.kwargs = kwargs

    def __str__(self):
        return repr(self)

    def __repr__(self):
        r = 'ioHubError:\nArgs: {0}\n'.format(self.args)
        for k, v in self.kwargs.items():
            r += '\t{0}: {1}\n'.format(k, v)
        return r



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""This stimulus class defines a field of dots with an update rule that
determines how they change on every call to the .draw() method.
"""

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2020 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

# Bugfix by Andrew Schofield.
# Replaces out of bounds but still live dots at opposite edge of aperture 
# instead of randomly within the field. This stops the concentration of dots at 
# one side of field when lifetime is long.
# Update the dot direction immediately for 'walk' as otherwise when the 
# coherence varies some signal dots will inherit the random directions of 
# previous walking dots.
# Provide a visible wrapper function to refresh all the dot locations so that 
# the whole field can be more easily refreshed between trials.

from __future__ import absolute_import, division, print_function

from builtins import str
from builtins import range

# Ensure setting pyglet.options['debug_gl'] to False is done prior to any
# other calls to pyglet or pyglet submodules, otherwise it may not get picked
# up by the pyglet GL engine and have no effect.
# Shaders will work but require OpenGL2.0 drivers AND PyOpenGL3.0+
import pyglet
pyglet.options['debug_gl'] = False
import ctypes
GL = pyglet.gl

import psychopy  # so we can get the __path__
from psychopy import logging

# tools must only be imported *after* event or MovieStim breaks on win32
# (JWP has no idea why!)
from psychopy.tools.attributetools import attributeSetter, setAttribute
from psychopy.tools.arraytools import val2array
from psychopy.visual.basevisual import (BaseVisualStim, ColorMixin,
                                        ContainerMixin)

import numpy as np

# some constants
_piOver2 = np.pi / 2.
_piOver180 = np.pi / 180.
_2pi = 2 * np.pi


class DotStim(BaseVisualStim, ColorMixin, ContainerMixin):
    """This stimulus class defines a field of dots with an update rule that
    determines how they change on every call to the .draw() method.

    This single class can be used to generate a wide variety of dot motion
    types. For a review of possible types and their pros and cons see Scase,
    Braddick & Raymond (1996). All six possible motions they describe can be
    generated with appropriate choices of the `signalDots` (which determines
    whether signal dots are the 'same' or 'different' on each frame),
    `noiseDots` (which determines the locations of the noise dots on each frame)
    and the `dotLife` (which determines for how many frames the dot will
    continue before being regenerated).

    The default settings (as of v1.70.00) is for the noise dots to have
    identical velocity but random direction and signal dots remain the 'same'
    (once a signal dot, always a signal dot).

    For further detail about the different configurations see :ref:`dots` in the
    Builder Components section of the documentation.

    If further customisation is required, then the DotStim should be subclassed
    and its _update_dotsXY and _newDotsXY methods overridden.

    The maximum number of dots that can be drawn is limited by system
    performance.

    Attributes
    ----------
    fieldShape : str
        *'sqr'* or 'circle'. Defines the envelope used to present the dots. If
        changed while drawing, dots outside new envelope will be respawned.
    dotSize : float
        Dot size specified in pixels (overridden if `element` is specified).
        :ref:`operations <attrib-operations>` are supported.
    dotLife : int
        Number of frames each dot lives for (-1=infinite). Dot lives are
        initiated randomly from a uniform distribution from 0 to dotLife. If
        changed while drawing, the lives of all dots will be randomly initiated
        again.
    signalDots : str
        If 'same' then the signal and noise dots are constant. If 'different'
        then the choice of which is signal and which is noise gets randomised on
        each frame. This corresponds to Scase et al's (1996) categories of RDK.
    noiseDots : str
        Determines the behaviour of the noise dots, taken directly from Scase et
        al's (1996) categories. For 'position', noise dots take a random
        position every frame. For 'direction' noise dots follow a random, but
        constant direction. For 'walk' noise dots vary their direction every
        frame, but keep a constant speed.
    element : object
        This can be any object that has a ``.draw()`` method and a
        ``.setPos([x,y])`` method (e.g. a GratingStim, TextStim...)!! DotStim
        assumes that the element uses pixels as units. ``None`` defaults to
        dots.
    fieldPos : array_like
        Specifying the location of the centre of the stimulus using a
        :ref:`x,y-pair <attrib-xy>`. See e.g. :class:`.ShapeStim` for more
        documentation/examples on how to set position.
        :ref:`operations <attrib-operations>` are supported.
    fieldSize : array_like
        Specifying the size of the field of dots using a
        :ref:`x,y-pair <attrib-xy>`. See e.g. :class:`.ShapeStim` for more
        documentation/examples on how to set position.
        :ref:`operations <attrib-operations>` are supported.
    coherence : float
        Change the coherence (%) of the DotStim. This will be rounded according
        to the number of dots in the stimulus.
    dir : float
        Direction of the coherent dots in degrees. :ref:`operations
        <attrib-operations>` are supported.
    speed : float
        Speed of the dots (in *units*/frame). :ref:`operations
        <attrib-operations>` are supported.

    """
    def __init__(self,
                 win,
                 units='',
                 nDots=1,
                 coherence=0.5,
                 fieldPos=(0.0, 0.0),
                 fieldSize=(1.0, 1.0),
                 fieldShape='sqr',
                 dotSize=2.0,
                 dotLife=3,
                 dir=0.0,
                 speed=0.5,
                 rgb=None,
                 color=(1.0, 1.0, 1.0),
                 colorSpace='rgb',
                 opacity=1.0,
                 contrast=1.0,
                 depth=0,
                 element=None,
                 signalDots='same',
                 noiseDots='direction',
                 name=None,
                 autoLog=None):
        """
        Parameters
        ----------
        win : window.Window
            Window this stimulus is associated with.
        units : str
            Units to use.
        nDots : int
            Number of dots to present in the field.
        coherence : float
            Proportion of dots which are coherent. This value can be set using
            the `coherence` property after initialization.
        fieldPos : array_like
            (x,y) or [x,y] position of the field. This value can be set using
            the `fieldPos` property after initialization.
        fieldSize : array_like, int or float
            (x,y) or [x,y] or single value (applied to both dimensions). Sizes
            can be negative and can extend beyond the window. This value can be
            set using the `fieldSize` property after initialization.
        fieldShape : str
            Defines the envelope used to present the dots. If changed while
            drawing by setting the `fieldShape` property, dots outside new
            envelope will be respawned., valid values are 'square', 'sqr' or
            'circle'.
        dotSize : array_like or float
            Size of the dots. If given an array, the sizes of individual dots
            will be set. The array must have length `nDots`. If a single value
            is given, all dots will be set to the same size.
        dotLife : int
            Lifetime of a dot in frames. Dot lives are initiated randomly from a
            uniform distribution from 0 to dotLife. If changed while drawing,
            the lives of all dots will be randomly initiated again. A value of
            -1 results in dots having an infinite lifetime. This value can be
            set using the `dotLife` property after initialization.
        dir : float
            Direction of the coherent dots in degrees. At 0 degrees, coherent
            dots will move from left to right. Increasing the angle will rotate
            the direction counter-clockwise. This value can be set using the
            `dir` property after initialization.
        speed : float
            Speed of the dots (in *units* per frame). This value can be set
            using the `speed` property after initialization.
        rgb : array_like, optional
            Color of the dots in form (r, g, b) or [r, g, b]. **Deprecated**,
            use `color` instead.
        color : array_like or str
            Color of the dots in form (r, g, b) or [r, g, b].
        colorSpace : str
            Colorspace to use.
        opacity : float
            Opacity of the dots from 0.0 to 1.0.
        contrast : float
            Contrast of the dots 0.0 to 1.0. This value is simply multiplied by
            the `color` value.
        depth : float
            **Deprecated**, depth is now controlled simply by drawing order.
        element : object
            This can be any object that has a ``.draw()`` method and a
            ``.setPos([x,y])`` method (e.g. a GratingStim, TextStim...)!!
            DotStim assumes that the element uses pixels as units.
            ``None`` defaults to dots.
        signalDots : str
            If 'same' then the signal and noise dots are constant. If different
            then the choice of which is signal and which is noise gets
            randomised on each frame. This corresponds to Scase et al's (1996)
            categories of RDK. This value can be set using the `signalDots`
            property after initialization.
        noiseDots : str
            Determines the behaviour of the noise dots, taken directly from
            Scase et al's (1996) categories. For 'position', noise dots take a
            random position every frame. For 'direction' noise dots follow a
            random, but constant direction. For 'walk' noise dots vary their
            direction every frame, but keep a constant speed. This value can be
            set using the `noiseDots` property after initialization.
        name : str, optional
            Optional name to use for logging.
        autoLog : bool
            Enable automatic logging.

        """
        # what local vars are defined (these are the init params) for use by
        # __repr__
        self._initParams = __builtins__['dir']()
        self._initParams.remove('self')

        super(DotStim, self).__init__(win, units=units, name=name,
                                      autoLog=False)  # set at end of init

        self.nDots = nDots
        # pos and size are ambiguous for dots so DotStim explicitly has
        # fieldPos = pos, fieldSize=size and then dotSize as additional param
        self.fieldPos = fieldPos  # self.pos is also set here
        self.fieldSize = val2array(fieldSize, False)  # self.size is also set
        if type(dotSize) in (tuple, list):
            self.dotSize = np.array(dotSize)
        else:
            self.dotSize = dotSize
        if self.win.useRetina:
            self.dotSize *= 2  # double dot size to make up for 1/2-size pixels
        self.fieldShape = fieldShape
        self.__dict__['dir'] = dir
        self.speed = speed
        self.element = element
        self.dotLife = dotLife
        self.signalDots = signalDots
        self.opacity = float(opacity)
        self.contrast = float(contrast)

        self.useShaders = False  # not needed for dots?
        self.colorSpace = colorSpace
        if rgb != None:
            logging.warning("Use of rgb arguments to stimuli are deprecated."
                            " Please use color and colorSpace args instead")
            self.setColor(rgb, colorSpace='rgb', log=False)
        else:
            self.setColor(color, log=False)

        self.depth = depth

        # initialise the dots themselves - give them all random dir and then
        # fix the first n in the array to have the direction specified
        self.coherence = coherence  # using the attributeSetter
        self.noiseDots = noiseDots

        # initialise a random array of X,Y
        self._verticesBase = self._dotsXY = self._newDotsXY(self.nDots)
        # all dots have the same speed
        self._dotsSpeed = np.ones(self.nDots, dtype=float) * self.speed
        # abs() means we can ignore the -1 case (no life)
        self._dotsLife = np.abs(dotLife) * np.random.rand(self.nDots)
        # pre-allocate array for flagging dead dots
        self._deadDots = np.zeros(self.nDots, dtype=bool)
        # set directions (only used when self.noiseDots='direction')
        self._dotsDir = np.random.rand(self.nDots) * _2pi
        self._dotsDir[self._signalDots] = self.dir * _piOver180

        self._update_dotsXY()

        # set autoLog now that params have been initialised
        wantLog = autoLog is None and self.win.autoLog
        self.__dict__['autoLog'] = autoLog or wantLog
        if self.autoLog:
            logging.exp("Created %s = %s" % (self.name, str(self)))

    def set(self, attrib, val, op='', log=None):
        """DEPRECATED: DotStim.set() is obsolete and may not be supported
        in future versions of PsychoPy. Use the specific method for each
        parameter instead (e.g. setFieldPos(), setCoherence()...).
        """
        self._set(attrib, val, op, log=log)

    @attributeSetter
    def fieldShape(self, fieldShape):
        """*'sqr'* or 'circle'. Defines the envelope used to present the dots.
        If changed while drawing, dots outside new envelope will be respawned.
        """
        self.__dict__['fieldShape'] = fieldShape

    @attributeSetter
    def dotSize(self, dotSize):
        """Float specified in pixels (overridden if `element` is specified).
        :ref:`operations <attrib-operations>` are supported."""
        self.__dict__['dotSize'] = dotSize

    @attributeSetter
    def dotLife(self, dotLife):
        """Int. Number of frames each dot lives for (-1=infinite).
        Dot lives are initiated randomly from a uniform distribution
        from 0 to dotLife. If changed while drawing, the lives of all
        dots will be randomly initiated again.

        :ref:`operations <attrib-operations>` are supported.
        """
        self.__dict__['dotLife'] = dotLife
        self._dotsLife = abs(self.dotLife) * np.random.rand(self.nDots)

    @attributeSetter
    def signalDots(self, signalDots):
        """str - 'same' or *'different'*
        If 'same' then the signal and noise dots are constant. If different
        then the choice of which is signal and which is noise gets
        randomised on each frame. This corresponds to Scase et al's (1996)
        categories of RDK.
        """
        self.__dict__['signalDots'] = signalDots

    @attributeSetter
    def noiseDots(self, noiseDots):
        """str - *'direction'*, 'position' or 'walk'
        Determines the behaviour of the noise dots, taken directly from
        Scase et al's (1996) categories. For 'position', noise dots take a
        random position every frame. For 'direction' noise dots follow a
        random, but constant direction. For 'walk' noise dots vary their
        direction every frame, but keep a constant speed.
        """
        self.__dict__['noiseDots'] = noiseDots
        self.coherence = self.coherence  # update using attributeSetter

    @attributeSetter
    def element(self, element):
        """*None* or a visual stimulus object
        This can be any object that has a ``.draw()`` method and a
        ``.setPos([x,y])`` method (e.g. a GratingStim, TextStim...)!!
        DotStim assumes that the element uses pixels as units.
        ``None`` defaults to dots.

        See `ElementArrayStim` for a faster implementation of this idea.
        """
        self.__dict__['element'] = element

    @attributeSetter
    def fieldPos(self, pos):
        """Specifying the location of the centre of the stimulus
        using a :ref:`x,y-pair <attrib-xy>`.
        See e.g. :class:`.ShapeStim` for more documentation / examples
        on how to set position.

        :ref:`operations <attrib-operations>` are supported.
        """
        # Isn't there a way to use BaseVisualStim.pos.__doc__ as docstring
        # here?
        self.pos = pos  # using BaseVisualStim. we'll store this as both
        self.__dict__['fieldPos'] = self.pos

    def setFieldPos(self, val, op='', log=None):
        """Usually you can use 'stim.attribute = value' syntax instead, but use 
        this method if you need to suppress the log message.
        """
        setAttribute(self, 'fieldPos', val, log, op)  # calls attributeSetter

    def setPos(self, newPos=None, operation='', units=None, log=None):
        """Obsolete - users should use setFieldPos instead of setPos
        """
        logging.error("User called DotStim.setPos(pos). "
                      "Use DotStim.SetFieldPos(pos) instead.")

    def setFieldSize(self, val, op='', log=None):
        """Usually you can use 'stim.attribute = value' syntax instead, but use 
        this method if you need to suppress the log message.
        """
        setAttribute(self, 'fieldSize', val, log, op)  # calls attributeSetter

    @attributeSetter
    def fieldSize(self, size):
        """Specifying the size of the field of dots using a
        :ref:`x,y-pair <attrib-xy>`. See e.g. :class:`.ShapeStim` for more 
        documentation/examples on how to set position.

        :ref:`operations <attrib-operations>` are supported.
        """
        # Isn't there a way to use BaseVisualStim.pos.__doc__ as docstring
        # here?
        self.size = size  # using BaseVisualStim. we'll store this as both
        self.__dict__['fieldSize'] = self.size

    @attributeSetter
    def coherence(self, coherence):
        """Scalar between 0 and 1.

        Change the coherence (%) of the DotStim. This will be rounded according 
        to the number of dots in the stimulus.

        :ref:`operations <attrib-operations>` are supported.
        """
        if not 0 <= coherence <= 1:
            raise ValueError('DotStim.coherence must be between 0 and 1')

        _cohDots = coherence * self.nDots

        self.__dict__['coherence'] = round(_cohDots) /self.nDots
        self._signalDots = np.zeros(self.nDots, dtype=bool)
        self._signalDots[0:int(self.coherence * self.nDots)] = True
        # for 'direction' method we need to update the direction of the number
        # of signal dots immediately, but for other methods it will be done
        # during updateXY

        # NB - AJS Actually you need to do this for 'walk' also
        # otherwise would be signal dots adopt random directions when the become
        # sinal dots in later trails
        if self.noiseDots in ('direction', 'position', 'walk'):
            self._dotsDir = np.random.rand(self.nDots) * _2pi
            self._dotsDir[self._signalDots] = self.dir * _piOver180

    def setFieldCoherence(self, val, op='', log=None):
        """Usually you can use 'stim.attribute = value' syntax instead, but use 
        this method if you need to suppress the log message.
        """
        setAttribute(self, 'coherence', val, log, op)  # calls attributeSetter

    @attributeSetter
    def dir(self, dir):
        """float (degrees). direction of the coherent dots. :ref:`operations 
        <attrib-operations>` are supported.
        """
        # check which dots are signal before setting new dir
        signalDots = self._dotsDir == (self.dir * _piOver180)
        self.__dict__['dir'] = dir

        # dots currently moving in the signal direction also need to update
        # their direction
        self._dotsDir[signalDots] = self.dir * _piOver180

    def setDir(self, val, op='', log=None):
        """Usually you can use 'stim.attribute = value' syntax instead, but use 
        this method if you need to suppress the log message.
        """
        setAttribute(self, 'dir', val, log, op)

    @attributeSetter
    def speed(self, speed):
        """float. speed of the dots (in *units*/frame). :ref:`operations 
        <attrib-operations>` are supported.
        """
        self.__dict__['speed'] = speed

    def setSpeed(self, val, op='', log=None):
        """Usually you can use 'stim.attribute = value' syntax instead, but use 
        this method if you need to suppress the log message.
        
        """
        setAttribute(self, 'speed', val, log, op)

    def draw(self, win=None):
        """Draw the stimulus in its relevant window. You must call this method
        after every MyWin.flip() if you want the stimulus to appear on that
        frame and then update the screen again.

        Parameters
        ----------
        win : window.Window, optional
            Window to draw dots to. If `None`, dots will be drawn to the parent
            window.

        """
        if win is None:
            win = self.win
        self._selectWindow(win)

        self._update_dotsXY()

        GL.glPushMatrix()  # push before drawing, pop after

        # draw the dots
        if self.element is None:
            win.setScale('pix')
            GL.glPointSize(self.dotSize)

            # load Null textures into multitexteureARB - they modulate with
            # glColor
            GL.glActiveTexture(GL.GL_TEXTURE0)
            GL.glEnable(GL.GL_TEXTURE_2D)
            GL.glBindTexture(GL.GL_TEXTURE_2D, 0)
            GL.glActiveTexture(GL.GL_TEXTURE1)
            GL.glEnable(GL.GL_TEXTURE_2D)
            GL.glBindTexture(GL.GL_TEXTURE_2D, 0)

            CPCD = ctypes.POINTER(ctypes.c_double)
            GL.glVertexPointer(2, GL.GL_DOUBLE, 0,
                               self.verticesPix.ctypes.data_as(CPCD))
            desiredRGB = self._getDesiredRGB(self.rgb, self.colorSpace,
                                             self.contrast)

            GL.glColor4f(desiredRGB[0], desiredRGB[1], desiredRGB[2],
                         self.opacity)
            GL.glEnableClientState(GL.GL_VERTEX_ARRAY)
            GL.glDrawArrays(GL.GL_POINTS, 0, self.nDots)
            GL.glDisableClientState(GL.GL_VERTEX_ARRAY)
        else:
            # we don't want to do the screen scaling twice so for each dot
            # subtract the screen centre
            initialDepth = self.element.depth
            for pointN in range(0, self.nDots):
                _p = self.verticesPix[pointN, :] + self.fieldPos
                self.element.setPos(_p)
                self.element.draw()
            # reset depth before going to next frame
            self.element.setDepth(initialDepth)
        GL.glPopMatrix()

    def _newDotsXY(self, nDots):
        """Returns a uniform spread of dots, according to the `fieldShape` and
        `fieldSize`.

        Parameters
        ----------
        nDots : int
            Number of dots to sample.

        Returns
        -------
        ndarray
            Nx2 array of X and Y positions of dots.

        Examples
        --------
        Create a new array of dot positions::

            dots = self._newDots(nDots)

        """
        if self.fieldShape == 'circle':
            length = np.sqrt(np.random.uniform(0, 1, (nDots,)))
            angle = np.random.uniform(0., _2pi, (nDots,))

            newDots = np.zeros((nDots, 2))
            newDots[:, 0] = length * np.cos(angle)
            newDots[:, 1] = length * np.sin(angle)

            newDots *= self.fieldSize * .5
        else:
            newDots = np.random.uniform(-0.5, 0.5, size = (nDots, 2)) * self.fieldSize

        return newDots

    def refreshDots(self):
        """Callable user function to choose a new set of dots."""
        self._verticesBase = self._dotsXY = self._newDotsXY(self.nDots)

        # Don't allocate another array if the new number of dots is equal to
        # the last.
        if self.nDots != len(self._deadDots):
            self._deadDots = np.zeros(self.nDots, dtype=bool)

    def _update_dotsXY(self):
        """The user shouldn't call this - its gets done within draw().
        """
        # Find dead dots, update positions, get new positions for
        # dead and out-of-bounds
        # renew dead dots
        if self.dotLife > 0:  # if less than zero ignore it
            # decrement. Then dots to be reborn will be negative
            self._dotsLife -= 1
            self._deadDots[:] = (self._dotsLife <= 0)
            self._dotsLife[self._deadDots] = self.dotLife
        else:
            self._deadDots[:] = False

        # update XY based on speed and dir
        # NB self._dotsDir is in radians, but self.dir is in degs
        # update which are the noise/signal dots
        if self.signalDots == 'different':
            #  **up to version 1.70.00 this was the other way around,
            # not in keeping with Scase et al**
            # noise and signal dots change identity constantly
            np.random.shuffle(self._dotsDir)
            # and then update _signalDots from that
            self._signalDots = (self._dotsDir == (self.dir * _piOver180))

        # update the locations of signal and noise; 0 radians=East!
        reshape = np.reshape
        if self.noiseDots == 'walk':
            # noise dots are ~self._signalDots
            sig = np.random.rand(np.sum(~self._signalDots))
            self._dotsDir[~self._signalDots] = sig * _2pi
            # then update all positions from dir*speed
            cosDots = reshape(np.cos(self._dotsDir), (self.nDots,))
            sinDots = reshape(np.sin(self._dotsDir), (self.nDots,))
            self._verticesBase[:, 0] += self.speed * cosDots
            self._verticesBase[:, 1] += self.speed * sinDots
        elif self.noiseDots == 'direction':
            # simply use the stored directions to update position
            cosDots = reshape(np.cos(self._dotsDir), (self.nDots,))
            sinDots = reshape(np.sin(self._dotsDir), (self.nDots,))
            self._verticesBase[:, 0] += self.speed * cosDots
            self._verticesBase[:, 1] += self.speed * sinDots
        elif self.noiseDots == 'position':
            # update signal dots
            sd = self._signalDots
            sdSum = self._signalDots.sum()
            cosDots = reshape(np.cos(self._dotsDir[sd]), (sdSum,))
            sinDots = reshape(np.sin(self._dotsDir[sd]), (sdSum,))
            self._verticesBase[sd, 0] += self.speed * cosDots
            self._verticesBase[sd, 1] += self.speed * sinDots
            # update noise dots
            self._deadDots[:] = self._deadDots + (~self._signalDots)

        # handle boundaries of the field
        if self.fieldShape in (None, 'square', 'sqr'):
            out0 = (np.abs(self._verticesBase[:, 0]) > .5 * self.fieldSize[0])
            out1 = (np.abs(self._verticesBase[:, 1]) > .5 * self.fieldSize[1])
            outofbounds = out0 + out1
        else:
            # transform to a normalised circle (radius = 1 all around)
            # then to polar coords to check
            # the normalised XY position (where radius should be < 1)
            normXY = self._verticesBase / .5 / self.fieldSize
            # add out-of-bounds to those that need replacing
            outofbounds = np.hypot(normXY[:, 0], normXY[:, 1]) > 1.

        # update any dead dots
        nDead = self._deadDots.sum()
        if nDead:
            self._verticesBase[self._deadDots, :] = self._newDotsXY(nDead)

        # Reposition any dots that have gone out of bounds. Net effect is to
        # place dot one step inside the boundary on the other side of the
        # aperture.
        nOutOfBounds = outofbounds.sum()
        if nOutOfBounds:
            self._verticesBase[outofbounds, :] = self._newDotsXY(nOutOfBounds)

        # update the pixel XY coordinates in pixels (using _BaseVisual class)
        self._updateVertices()
